#!/usr/bin/env python3
#
# Copyright © 2020 Red Hat, Inc.
# SPDX-License-Identifier: MIT

import concurrent.futures
import itertools
import logging
import os
import resource
import subprocess
import sys
import tempfile
import unittest
from abc import ABCMeta, abstractmethod
from collections.abc import Callable, Generator
from dataclasses import dataclass
from functools import reduce
from pathlib import Path
from typing import ClassVar

try:
    top_builddir = Path(os.environ["top_builddir"])
    top_srcdir = Path(os.environ["top_srcdir"])
except KeyError:
    print(
        "Required environment variables not found: top_srcdir/top_builddir",
        file=sys.stderr,
    )

    top_srcdir = Path(".")
    try:
        top_builddir = next(Path(".").glob("**/meson-logs/")).parent
    except StopIteration:
        sys.exit(1)
    print(
        'Using srcdir "{}", builddir "{}"'.format(top_srcdir, top_builddir),
        file=sys.stderr,
    )

TIMEOUT = 10.0  # seconds

# Unset some environment variables, so that testing --enable-environment-names
# does not actually depend on the current environment.
for key in (
    "XKB_DEFAULT_RULES",
    "XKB_DEFAULT_MODEL",
    "XKB_DEFAULT_LAYOUT",
    "XKB_DEFAULT_VARIANT",
    "XKB_DEFAULT_OPTIONS",
):
    if key in os.environ:
        del os.environ[key]

# Ensure locale is C, so we can check error messages in English
os.environ["LC_ALL"] = "C.UTF-8"

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger("test")
logger.setLevel(logging.DEBUG)


def powerset(iterable):
    "Subsequences of the iterable from shortest to longest."
    # powerset([1,2,3]) → () (1,) (2,) (3,) (1,2) (1,3) (2,3) (1,2,3)
    s = tuple(iterable)
    return itertools.chain.from_iterable(
        itertools.combinations(s, r) for r in range(len(s) + 1)
    )


# Permutation of RMLVO that we use in multiple tests
rmlvo_options = (
    "--rules=evdev",
    "--model=pc104",
    "--layout=ch",
    "--options=eurosign:5",
    "--enable-environment-names",
)
rmlvos = tuple(
    map(
        list,
        itertools.chain(
            itertools.permutations(rmlvo_options[:-1]),
            powerset(rmlvo_options),
        ),
    )
)


@dataclass
class RMLVO:
    rules: str | None = None
    model: str | None = None
    layout: str | None = None
    variant: str | None = None
    options: str | None = None
    env: bool = False

    def __iter__(self) -> Generator[tuple[str, ...]]:
        if self.rules is not None:
            yield "--rules", self.rules
        if self.model is not None:
            yield "--model", self.model
        if self.layout is not None:
            yield "--layout", self.layout
        if self.variant is not None:
            yield "--variant", self.variant
        if self.options is not None:
            yield "--options", self.options
        if self.env:
            yield ("--enable-environment-names",)

    @property
    def args(self) -> list[str]:
        return list(itertools.chain.from_iterable(self))


class Target(metaclass=ABCMeta):
    @property
    @abstractmethod
    def args(self) -> list[str]: ...

    @property
    def stdin(self) -> str | None:
        return None


class RmlvoTarget(Target):
    @property
    def args(self) -> list[str]:
        return ["--rmlvo"]


class KccgstTarget(Target):
    @property
    def args(self) -> list[str]:
        return ["--kccgst"]


class KccgstYamlTarget(Target):
    @property
    def args(self) -> list[str]:
        return ["--kccgst-yaml"]


@dataclass
class KeymapTarget(Target, RMLVO):
    arg: bool = False
    path: Path | None = None
    stdin: str | None = None

    def _args(self) -> Generator[str]:
        yield from itertools.chain.from_iterable(self)
        if self.arg:
            yield "--keymap"
        if self.path:
            yield str(self.path)

    @property
    def args(self) -> list[str]:
        return list(self._args())

    @property
    def use_rmlvo(self) -> bool:
        return (
            not self.arg
            and not self.path
            and (
                bool(self.rules)
                or bool(self.model)
                or bool(self.layout)
                or bool(self.variant)
                or bool(self.options)
                or bool(self.env)
                or not self.stdin
            )
        )


def _disable_coredump():
    resource.setrlimit(resource.RLIMIT_CORE, (0, 0))


def run_command(args, input: str | None = None) -> tuple[int, str, str]:
    logger.debug("run command: {}".format(" ".join(args)))

    try:
        p = subprocess.run(
            args,
            preexec_fn=_disable_coredump,
            capture_output=True,
            input=input,
            text=True,
            timeout=0.7,
        )
        return p.returncode, p.stdout, p.stderr
    except subprocess.TimeoutExpired as e:
        return (
            0,
            e.stdout.decode("utf-8") if e.stdout else "",
            e.stderr.decode("utf-8") if e.stderr else "",
        )


@dataclass
class XkbcliTool:
    xkbcli_tool: ClassVar[str] = "xkbcli"
    tool_path: ClassVar[Path] = top_builddir

    subtool: str | None = None
    skipIf: tuple[tuple[bool, str], ...] = ()
    skipError: tuple[tuple[Callable[[int, str, str], bool], str], ...] = ()

    def run_command(self, args, input: str | None = None) -> tuple[int, str, str]:
        for condition, reason in self.skipIf:
            if condition:
                raise unittest.SkipTest(reason)
        if self.subtool is not None:
            tool = "{}-{}".format(self.xkbcli_tool, self.subtool)
        else:
            tool = self.xkbcli_tool
        args = [os.path.join(self.tool_path, tool)] + args

        return run_command(args, input=input)

    def run_command_success(self, args, input: str | None = None) -> tuple[str, str]:
        rc, stdout, stderr = self.run_command(args, input=input)
        if rc != 0:
            for testfunc, reason in self.skipError:
                if testfunc(rc, stdout, stderr):
                    raise unittest.SkipTest(reason)
        assert rc == 0, (rc, stdout, stderr)
        return stdout, stderr

    def run_command_invalid(
        self, args, input: str | None = None
    ) -> tuple[int, str, str]:
        rc, stdout, stderr = self.run_command(args, input=input)
        assert rc == 2, (rc, stdout, stderr)
        return rc, stdout, stderr

    def run_command_unrecognized_option(self, args):
        rc, stdout, stderr = self.run_command(args)
        assert rc == 2, (rc, stdout, stderr)
        assert stdout.startswith("Usage") or stdout == ""
        assert "unrecognized option" in stderr

    def run_command_missing_arg(self, args):
        rc, stdout, stderr = self.run_command(args)
        assert rc == 2, (rc, stdout, stderr)
        assert stdout.startswith("Usage") or stdout == ""
        assert "requires an argument" in stderr

    def __str__(self):
        return str(self.subtool)


class TestXkbcli(unittest.TestCase):
    xkbcli: ClassVar[XkbcliTool]
    xkbcli_list: ClassVar[XkbcliTool]
    xkbcli_how_to_type: ClassVar[XkbcliTool]
    xkbcli_compile_keymap: ClassVar[XkbcliTool]
    xkbcli_compile_compose: ClassVar[XkbcliTool]
    xkbcli_interactive_evdev: ClassVar[XkbcliTool]
    xkbcli_interactive_x11: ClassVar[XkbcliTool]
    xkbcli_interactive_wayland: ClassVar[XkbcliTool]
    xkbcli_interactive: ClassVar[XkbcliTool]
    xkbcli_dump_keymap_x11: ClassVar[XkbcliTool]
    xkbcli_dump_keymap_wayland: ClassVar[XkbcliTool]
    xkbcli_dump_keymap: ClassVar[XkbcliTool]
    all_tools: ClassVar[list[XkbcliTool]]

    @classmethod
    def setUpClass(cls):
        cls.xkbcli = XkbcliTool()
        cls.xkbcli_list = XkbcliTool(
            "list",
            skipIf=(
                (
                    not int(os.getenv("HAVE_XKBCLI_LIST", "1")),
                    "xkbregistory not enabled",
                ),
            ),
        )
        cls.xkbcli_how_to_type = XkbcliTool("how-to-type")
        cls.xkbcli_compile_keymap = XkbcliTool("compile-keymap")
        cls.xkbcli_compile_compose = XkbcliTool("compile-compose")
        no_interactive_evdev = (
            (
                not int(os.getenv("HAVE_XKBCLI_INTERACTIVE_EVDEV", "1")),
                "evdev not enabled",
            ),
            (not os.path.exists("/dev/input/event0"), "event node required"),
            (
                not os.access("/dev/input/event0", os.R_OK),
                "insufficient permissions",
            ),
        )
        cls.xkbcli_interactive_evdev = XkbcliTool(
            "interactive-evdev",
            skipIf=no_interactive_evdev,
            skipError=(
                (
                    lambda rc, stdout, stderr: "Couldn't find any keyboards" in stderr,
                    "No keyboards available",
                ),
            ),
        )
        no_interactive_x11 = (
            (
                not int(os.getenv("HAVE_XKBCLI_INTERACTIVE_X11", "1")),
                "x11 not enabled",
            ),
            (not os.getenv("DISPLAY"), "DISPLAY not set"),
        )
        cls.xkbcli_interactive_x11 = XkbcliTool(
            "interactive-x11",
            skipIf=no_interactive_x11,
        )
        no_interactive_wayland = (
            (
                not int(os.getenv("HAVE_XKBCLI_INTERACTIVE_WAYLAND", "1")),
                "wayland not enabled",
            ),
            (not os.getenv("WAYLAND_DISPLAY"), "WAYLAND_DISPLAY not set"),
        )
        cls.xkbcli_interactive_wayland = XkbcliTool(
            "interactive-wayland",
            skipIf=no_interactive_wayland,
        )
        # NOTE: `interactive` cannot be tested, because it hardcodes the paths
        #        of the `xkbcli-interactive-{wayland,x11}` tools it calls to the
        #        *install* directory, while we need to use those of the *build*
        #        directory.

        cls.xkbcli_dump_keymap_x11 = XkbcliTool(
            "dump-keymap-x11",
            skipIf=no_interactive_x11,
        )
        cls.xkbcli_dump_keymap_wayland = XkbcliTool(
            "dump-keymap-wayland",
            skipIf=no_interactive_wayland,
        )
        # NOTE: `dump-keymap` cannot be tested, because it hardcodes the paths
        #        of the `xkbcli-dump-keymap-{wayland,x11}` tools it calls to the
        #        *install* directory, while we need to use those of the *build*
        #        directory.

        cls.all_tools = [
            cls.xkbcli,
            cls.xkbcli_list,
            cls.xkbcli_how_to_type,
            cls.xkbcli_compile_keymap,
            cls.xkbcli_compile_compose,
            cls.xkbcli_interactive_evdev,
            cls.xkbcli_interactive_x11,
            cls.xkbcli_interactive_wayland,
            cls.xkbcli_dump_keymap_x11,
            cls.xkbcli_dump_keymap_wayland,
        ]

    def test_help(self):
        # --help is supported by all tools
        for tool in self.all_tools:
            with self.subTest(tool=tool):
                stdout, stderr = tool.run_command_success(["--help"])
                assert stdout.startswith("Usage:")
                assert stderr == ""

    def test_invalid_option(self):
        # --foobar generates "Usage:" for all tools
        for tool in self.all_tools:
            with self.subTest(tool=tool):
                tool.run_command_unrecognized_option(["--foobar"])

    def test_xkbcli_version(self):
        # xkbcli --version
        stdout, stderr = self.xkbcli.run_command_success(["--version"])
        assert stdout.startswith("1")
        assert stderr == ""

    def test_xkbcli_too_many_args(self):
        self.xkbcli.run_command_invalid(["a"] * 64)

    def test_compile_keymap_args(self):
        xkb_root = Path(os.environ["XKB_CONFIG_ROOT"])
        keymap_path = xkb_root / "keymaps/masks.xkb"
        keymap_string = keymap_path.read_text(encoding="utf-8")
        test_option = "--test"
        target: Target
        for target in (
            RmlvoTarget(),
            KccgstTarget(),
            KccgstYamlTarget(),
            # Keymap from RMLVO
            KeymapTarget(),
            # Keymap from RMLVO (stdin ignored)
            KeymapTarget(layout="us", stdin=keymap_string),
            KeymapTarget(env=True, stdin=keymap_string),
            # Keymap parsed from keymap file
            KeymapTarget(arg=False, path=keymap_path),
            KeymapTarget(arg=True, path=keymap_path),
            # Keymap parsed from stdin
            KeymapTarget(arg=False, stdin=keymap_string),
            KeymapTarget(arg=True, stdin=keymap_string),
        ):
            for options in powerset(("--verbose", test_option)):
                args = list(options) + target.args
                with self.subTest(args=args):
                    stdout, stderr = self.xkbcli_compile_keymap.run_command_success(
                        args, input=target.stdin
                    )
                    if test_option in options:
                        # Expect no output
                        assert stdout == "", (target, stdout)
                    if isinstance(target, KeymapTarget) and test_option not in options:
                        # Check keymap output for excepted bits:
                        # - <AB01> is not defined in mask.xkb
                        # - virtual_modifiers Test01 is only defined in masks.xkb
                        assert ("<AB01>" in stdout) ^ (not target.use_rmlvo), (
                            target,
                            stdout,
                            stderr,
                        )
                        assert ("virtual_modifiers Test01" in stdout) ^ (
                            target.use_rmlvo
                        ), (target, stdout, stderr)

    def test_compile_keymap_mutually_exclusive_args(self):
        xkb_root = Path(os.environ["XKB_CONFIG_ROOT"])
        keymap_path = xkb_root / "keymaps/basic.xkb"
        keymap_string = keymap_path.read_text(encoding="utf-8")
        keymap_from_stdin = KeymapTarget(arg=True, stdin=keymap_string)
        keymap_from_path1 = KeymapTarget(arg=True, path=Path(keymap_path))
        keymap_from_path2 = KeymapTarget(arg=False, path=Path(keymap_path))
        rmlvo = RmlvoTarget()
        kccgst = KccgstTarget()
        kccgstYaml = KccgstYamlTarget()
        for entry in (
            # --keymap does not use RMLVO options
            ("--rules", "some-rules", keymap_from_stdin),
            ("--model", "some-model", keymap_from_stdin),
            ("--layout", "some-layout", keymap_from_stdin),
            ("--variant", "some-variant", keymap_from_stdin),
            ("--options", "some-option", keymap_from_stdin),
            ("--rules", "some-rules", keymap_from_path1),
            ("--model", "some-model", keymap_from_path1),
            ("--layout", "some-layout", keymap_from_path1),
            ("--variant", "some-variant", keymap_from_path1),
            ("--options", "some-option", keymap_from_path1),
            # Trailing keymap file with RMLVO options
            ("--rules", "some-rules", keymap_from_path2),
            ("--model", "some-model", keymap_from_path2),
            ("--layout", "some-layout", keymap_from_path2),
            ("--variant", "some-variant", keymap_from_path2),
            ("--options", "some-option", keymap_from_path2),
            # Incompatible output types
            (rmlvo, keymap_from_stdin),
            (rmlvo, keymap_from_path1),
            (rmlvo, keymap_from_path2),
            (kccgst, kccgstYaml),
            (kccgst, keymap_from_stdin),
            (kccgst, keymap_from_path1),
            (kccgst, keymap_from_path2),
            (kccgst, rmlvo),
            (kccgst, rmlvo, keymap_from_stdin),
            (kccgst, rmlvo, keymap_from_path1),
            (kccgst, rmlvo, keymap_from_path2),
            (kccgstYaml, keymap_from_stdin),
            (kccgstYaml, keymap_from_path1),
            (kccgstYaml, keymap_from_path2),
            (kccgstYaml, rmlvo),
            (kccgstYaml, rmlvo, keymap_from_stdin),
            (kccgstYaml, rmlvo, keymap_from_path1),
            (kccgstYaml, rmlvo, keymap_from_path2),
        ):
            with self.subTest(args=entry):
                args: list[str] = list(
                    itertools.chain.from_iterable(
                        arg.args if isinstance(arg, Target) else arg for arg in entry
                    )
                )
                input: str | None = reduce(
                    lambda acc, arg: acc
                    or (arg.stdin if isinstance(arg, Target) else None),
                    args,
                    None,
                )
                self.xkbcli_compile_keymap.run_command_invalid(args, input=input)

    def test_compile_keymap_rmlvo(self):
        def run(target, rmlvo):
            return self.xkbcli_compile_keymap.run_command_success(target + rmlvo)

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(run, target.args, rmlvo): (target, rmlvo)
                for target in (
                    RmlvoTarget(),
                    KccgstTarget(),
                    KccgstYamlTarget(),
                    KeymapTarget(),
                )
                for rmlvo in rmlvos
            }
            for future in concurrent.futures.as_completed(futures, TIMEOUT):
                target, rmlvo = futures[future]
                with self.subTest(target=target, rmlvo=rmlvo):
                    future.result()

    def test_compile_keymap_include(self):
        for args in (
            ["--include", ".", "--include-defaults"],
            ["--include", "/tmp", "--include-defaults"],
        ):
            with self.subTest(args=args):
                # Succeeds thanks to include-defaults
                self.xkbcli_compile_keymap.run_command_success(args)

    def test_compile_keymap_include_invalid(self):
        # A non-directory is rejected by default
        args = ["--include", "/proc/version"]
        rc, stdout, stderr = self.xkbcli_compile_keymap.run_command(args)
        assert rc == 1, (stdout, stderr)
        assert "There are no include paths to search" in stderr

        # A non-existing directory is rejected by default
        args = ["--include", "/tmp/does/not/exist"]
        rc, stdout, stderr = self.xkbcli_compile_keymap.run_command(args)
        assert rc == 1, (stdout, stderr)
        assert "There are no include paths to search" in stderr

        # Valid dir, but missing files
        args = ["--include", "/tmp"]
        rc, stdout, stderr = self.xkbcli_compile_keymap.run_command(args)
        assert rc == 1, (stdout, stderr)
        assert "Couldn't look up rules" in stderr

    def test_compile_compose(self):
        for args in (["--verbose"],):
            with self.subTest(args=args):
                self.xkbcli_compile_compose.run_command_success(args)

    def test_how_to_type(self):
        for args in (["--verbose", "1"],):
            with self.subTest(args=args):
                self.xkbcli_how_to_type.run_command_success(args)

        @dataclass
        class Entry:
            args: list[str]
            name: str
            value: int

        for entry in (
            # Unicode codepoint conversions, we support whatever strtol does
            Entry(args=["123"], name="braceleft", value=0x007B),
            Entry(args=["0123"], name="braceleft", value=0x007B),
            Entry(args=["0a"], name="Linefeed", value=0xFF0A),
            Entry(args=["0x123"], name="gcedilla", value=0x03BB),
            Entry(args=["U+123"], name="gcedilla", value=0x03BB),
            # Characters
            Entry(args=["1"], name="1", value=0x0031),
            Entry(args=["a"], name="a", value=0x0061),
            Entry(args=["á"], name="aacute", value=0x00E1),
            # Keysyms names (fallback without --keysym option)
            Entry(args=["acute"], name="acute", value=0x00B4),
            Entry(args=["U123"], name="U0123", value=0x1000123),
            # Keysyms names (with --keysym)
            Entry(args=["--keysym", "1"], name="1", value=0x0031),
            Entry(args=["--keysym", "a"], name="a", value=0x0061),
            Entry(args=["--keysym", "acute"], name="acute", value=0x00B4),
            Entry(args=["--keysym", "U123"], name="U0123", value=0x1000123),
            # Keysym values
            Entry(args=["--keysym", "123"], name="braceleft", value=0x007B),
            Entry(args=["--keysym", "0x123"], name="0x00000123", value=0x0123),
        ):
            with self.subTest(args=args):
                stdout, _stderr = self.xkbcli_how_to_type.run_command_success(
                    entry.args
                )
                expected = f"keysym: {entry.name} (0x{entry.value:04x})"
                lines = stdout.splitlines()
                assert len(lines) >= 1
                assert lines[0] == expected, (
                    entry,
                    f'expected: "{expected}", but got: "{lines[0]}"',
                )

    def test_how_to_type_rmlvo(self):
        def run(rmlvo):
            args = rmlvo + ["0x1234"]
            return self.xkbcli_how_to_type.run_command_success(args)

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = {executor.submit(run, rmlvo): rmlvo for rmlvo in rmlvos}
            for future in concurrent.futures.as_completed(futures, TIMEOUT):
                rmlvo = futures[future]
                with self.subTest(rmlvo=rmlvo):
                    future.result()

    def test_list_rmlvo(self):
        for args in (
            ["--verbose"],
            ["-v"],
            ["--verbose", "--load-exotic"],
            ["--load-exotic"],
            ["--ruleset=evdev"],
            ["--ruleset=base"],
        ):
            with self.subTest(args=args):
                self.xkbcli_list.run_command_success(args)

    def test_list_rmlvo_includes(self):
        args = ["/tmp/"]
        self.xkbcli_list.run_command_success(args)

    def test_list_rmlvo_includes_invalid(self):
        args = ["/proc/version"]
        rc, stdout, stderr = self.xkbcli_list.run_command(args)
        assert rc == 1
        assert "Failed to append include path" in stderr

    def test_list_rmlvo_includes_no_defaults(self):
        args = ["--skip-default-paths", "/tmp"]
        rc, stdout, stderr = self.xkbcli_list.run_command(args)
        assert rc == 1
        assert "Failed to parse XKB description" in stderr

    def test_interactive_evdev_rmlvo(self):
        def run(rmlvo):
            return self.xkbcli_interactive_evdev.run_command_success(rmlvo)

        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = {executor.submit(run, rmlvo): rmlvo for rmlvo in rmlvos}
            for future in concurrent.futures.as_completed(futures, TIMEOUT):
                rmlvo = futures[future]
                with self.subTest(rmlvo=rmlvo):
                    future.result()

    def test_interactive_evdev(self):
        # Note: --enable-compose fails if $prefix doesn't have the compose tables
        # installed
        for args in (
            ["--verbose"],
            ["--uniline"],
            ["--multiline"],
            ["--report-state-changes"],
            ["--enable-compose"],
            ["--consumed-mode=xkb"],
            ["--consumed-mode=gtk"],
            ["--without-x11-offset"],
        ):
            with self.subTest(args=args):
                self.xkbcli_interactive_evdev.run_command_success(args)

    def test_interactive_x11(self):
        for args in (
            ["--verbose"],
            ["--uniline"],
            ["--multiline"],
        ):
            with self.subTest(args=args):
                self.xkbcli_interactive_x11.run_command_success(args)

    def test_interactive_wayland(self):
        for args in (
            ["--verbose"],
            ["--uniline"],
            ["--multiline"],
        ):
            with self.subTest(args=args):
                self.xkbcli_interactive_wayland.run_command_success(args)


if __name__ == "__main__":
    with tempfile.TemporaryDirectory() as tmpdir:
        # Use our own test xkeyboard-config copy.
        os.environ["XKB_CONFIG_ROOT"] = str(top_srcdir / "test/data")
        # Use our own X11 locale copy.
        os.environ["XLOCALEDIR"] = str(top_srcdir / "test/data/locale")
        # Use our own locale.
        os.environ["LC_CTYPE"] = "en_US.UTF-8"
        # libxkbcommon has fallbacks when XDG_CONFIG_HOME isn't set so we need
        # to override it with a known (empty) directory. Otherwise our test
        # behavior depends on the system the test is run on.
        os.environ["XDG_CONFIG_HOME"] = tmpdir
        # Prevent the legacy $HOME/.xkb from kicking in.
        del os.environ["HOME"]
        # This needs to be separated if we do specific extra path testing
        os.environ["XKB_CONFIG_EXTRA_PATH"] = tmpdir
        os.environ["XKB_CONFIG_VERSIONED_EXTENSIONS_PATH"] = tmpdir
        os.environ["XKB_CONFIG_UNVERSIONED_EXTENSIONS_PATH"] = tmpdir

        unittest.main()



#!/usr/bin/env python3

# Copyright © 2025 Pierre Le Marre <dev@wismill.eu>
# SPDX-License-Identifier: MIT

"""
This script generate tests for merge modes.
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
import argparse
from collections import defaultdict
from collections.abc import Callable, Generator, Iterable, Iterator, Sequence
import dataclasses
import itertools
import re
import textwrap
from dataclasses import dataclass
from enum import Flag, IntFlag, StrEnum, auto, unique
from pathlib import Path
from string import Template
from typing import Any, ClassVar, NewType, Self

import jinja2

SCRIPT = Path(__file__)
# Root of the project
ROOT = SCRIPT.parent.parent

################################################################################
#
# XKB compiler
#
################################################################################


@dataclass
class XkbCompiler:
    name: str
    extended_syntax: bool

    default: ClassVar[str] = "xkbcommon"

    @property
    def is_default(self) -> bool:
        return self.name == self.default

    @property
    def suffix(self) -> str:
        return "" if self.is_default else "-" + self.name

    @classmethod
    def parse(cls, name: str) -> Self:
        match name:
            case cls.default:
                return cls(name=name, extended_syntax=True)
            case "xkbcomp":
                return cls(name=name, extended_syntax=False)
            case "kbvm":
                return cls(name=name, extended_syntax=True)
            case _:
                raise ValueError(f"Invalid XKB compiler: {name}")


################################################################################
#
# Merge modes
#
################################################################################


@unique
class MergeMode(StrEnum):
    Default = auto()
    Augment = auto()
    Override = auto()
    Replace = auto()

    @property
    def include(self) -> str:
        if self is self.__class__.Default:
            return "include"
        else:
            return self

    @property
    def char(self) -> str:
        match self:
            case self.__class__.Default:
                return "+"
            case self.__class__.Augment:
                return "|"
            case self.__class__.Override:
                return "+"
            case self.__class__.Replace:
                return "^"
            case _:
                raise ValueError(self)


INDENT = "\t"


################################################################################
#
# Components
#
################################################################################


@unique
class Component(StrEnum):
    keycodes = auto()
    types = auto()
    compat = auto()
    symbols = auto()

    def render_keymap(self, content) -> Generator[str]:
        yield "xkb_keymap {"
        for component in self.__class__:
            yield from (
                textwrap.indent(t, INDENT)
                for t in self.render(component, content if component is self else "")
            )
        yield "};"

    @classmethod
    def render(cls, component: Self, content: str, name: str = "") -> Generator[str]:
        if not content:
            match component:
                case cls.keycodes:
                    pass
                case cls.types:
                    pass
                case cls.compat:
                    content = "interpret.repeat= True;"
                case cls.symbols:
                    pass
        if name:
            name = f' "{name}"'
        if content:
            yield f'xkb_{component}{name} "" {{'
            yield textwrap.indent(content, INDENT)
            yield "};"
        else:
            yield f'xkb_{component}{name} "" {{}};'


class ComponentTemplate(Template):
    def __init__(self, template) -> Self:
        super().__init__(textwrap.dedent(template).strip())

    def __bool__(self) -> bool:
        return bool(self.template)

    def render(self, merge: MergeMode) -> str:
        mode = "" if merge is MergeMode.Default else f"{merge} "
        return self.substitute(mode=mode)

    def __add__(self, other: Any) -> Self:
        if isinstance(other, str):
            return self.__class__(self.template + "\n" + other)
        elif isinstance(other, Template):
            return self.__class__(self.template + "\n" + other.template)
        else:
            return NotImplemented

    def __radd__(self, other: Any) -> Self:
        if isinstance(other, str):
            return self.__class__(other + "\n" + self.template)
        elif isinstance(other, Template):
            return self.__class__(other.template + "\n" + self.template)
        else:
            return NotImplemented


################################################################################
#
# Test templates
#
################################################################################


@dataclass
class Test:
    title: str
    file: Path
    content: str
    expected: str


@dataclass
class TestGroup:
    title: str
    tests: Sequence[Test]

    c_template: ClassVar[Path] = Path("test/merge_modes.c.jinja")

    SLUG_PATTERN: ClassVar[re.Pattern[str]] = re.compile(r"[^-\w()]+")

    @classmethod
    def file(cls, title: str) -> Path:
        return Path(cls.SLUG_PATTERN.sub("-", title.lower().replace(" ", "_")))

    @classmethod
    def _write(
        cls,
        root: Path,
        jinja_env: jinja2.Environment,
        path: Path,
        **kwargs,
    ) -> None:
        template_path = root / cls.c_template
        template = jinja_env.get_template(template_path.relative_to(root).as_posix())

        with path.open("wt", encoding="utf-8") as fd:
            fd.writelines(template.generate(script=SCRIPT.relative_to(ROOT), **kwargs))

    @classmethod
    def write_c(
        cls,
        root: Path,
        jinja_env: jinja2.Environment,
        path: Path,
        compiler: XkbCompiler,
        tests: Sequence[TestGroup],
    ) -> None:
        return cls._write(
            root=root, jinja_env=jinja_env, path=path, compiler=compiler, tests=tests
        )

    @classmethod
    def write_data(
        cls,
        root: Path,
        compiler: XkbCompiler,
        tests: Sequence[TestGroup],
    ) -> None:
        tests_: dict[Path, list[Test]] = defaultdict(list)
        suffix = compiler.suffix
        for group in tests:
            for test in group.tests:
                file = test.file.parent
                file = file.with_name(file.name + suffix)
                tests_[file].append(test)
        for file, group_ in tests_.items():
            path = root / "test" / "data" / file
            with path.open("wt", encoding="utf-8") as fd:
                fd.write(
                    "// WARNING: This file was auto-generated by: "
                    f"{SCRIPT.relative_to(ROOT)}\n"
                )
                for test in group_:
                    fd.write(test.content)


@dataclass
class TestTemplate(metaclass=ABCMeta):
    title: str

    _plain: ClassVar[str] = "plain"
    _include: ClassVar[str] = "include"
    _base: ClassVar[str] = "base"
    _update: ClassVar[str] = "update"
    _file: ClassVar[str] = "merge_modes"

    def make_title(self, name: str) -> str:
        return f"{self.title}: {name}"

    @classmethod
    def make_file(cls, name: str) -> Path:
        return TestGroup.file(name)

    @abstractmethod
    def generate_tests(self) -> TestGroup: ...

    @abstractmethod
    def generate_data(self) -> TestGroup: ...

    @classmethod
    @abstractmethod
    def write(
        cls,
        root: Path,
        jinja_env: jinja2.Environment,
        c_file: Path | None,
        xkb: bool,
        compiler: XkbCompiler,
        tests: Sequence[Self],
        debug: bool,
    ): ...


@dataclass
class ComponentTestTemplate(TestTemplate):
    component: Component
    # Templates to test
    base_template: ComponentTemplate
    update_template: ComponentTemplate
    # Expected result
    augment: str
    override: str
    replace: str = ""

    def __post_init__(self):
        self.augment = textwrap.dedent(self.augment).strip()
        self.override = textwrap.dedent(self.override).strip()
        if not self.replace:
            self.replace = self.override
        else:
            self.replace = textwrap.dedent(self.replace).strip()

    def expected(self, mode: MergeMode) -> str:
        match mode:
            case MergeMode.Default:
                return self.override
            case MergeMode.Augment:
                return self.augment
            case MergeMode.Override:
                return self.override
            case MergeMode.Replace:
                return self.replace
            case _:
                raise ValueError(mode)

    def _make_section_name(self, type: str, mode: MergeMode) -> str:
        file = self.make_file(f"{self.title}-{mode}" if self.title else mode)
        return f"{type}-{file}"

    def _make_file(self, type: str, mode: MergeMode) -> Path:
        section_name = self._make_section_name(type, mode)
        return Path(self.component) / self._file / section_name

    def render_section(self, content: str, name: str = "") -> str:
        return "\n".join(self.component.render(self.component, content, name))

    def render_keymap(self, content: str) -> str:
        return "\n".join(self.component.render_keymap(content))

    def _generate_tests(
        self, render: Callable[[str], str], compiler: XkbCompiler
    ) -> Generator[Test]:
        include_suffix = compiler.suffix
        for base_mode, update_mode in itertools.product(MergeMode, MergeMode):
            # Plain
            content = (
                self.base_template.render(base_mode)
                + ("\n" if self.base_template else "")
                + self.update_template.render(update_mode)
            )
            yield Test(
                title=self.make_title(
                    f"{self._plain} ({base_mode}) - {self._plain} ({update_mode})"
                ),
                file=self.make_file(self.title) / self.make_file(update_mode),
                content=render(content),
                expected=render(self.expected(update_mode)),
            )

        for base_mode, update_mode in itertools.product(
            (MergeMode.Default,), MergeMode
        ):
            # Plain + Include
            file = self._make_file(self._update, update_mode)
            section = file.name
            file = file.parent.name
            for mode in MergeMode:
                yield Test(
                    title=self.make_title(
                        f"{self._plain} ({base_mode}) - {mode.include} ({update_mode})"
                    ),
                    file=self.make_file(self.title) / self.make_file(mode),
                    content=render(
                        self.base_template.render(base_mode)
                        + (
                            f'\n{mode.include} "{file}{include_suffix}({section})"'
                            if self.base_template
                            else ""
                        )
                    ),
                    # Include does not leak local merge modes
                    expected=render(self.expected(mode)),
                )

        # for base_mode, update_mode in itertools.product(MergeMode, MergeMode):
        #     # Include + Plain
        #     file = self._make_file(self.base, base_mode)
        #     section = file.name
        #     file = file.parent.name
        #     for mode in (MergeMode.default,):
        #         yield Test(
        #             title=self.make_title(
        #                 f"{mode.include} ({base_mode}) - {self.plain} ({update_mode})"
        #             ),
        #             file=self.make_file(self.title)
        #             / self.make_file(
        #                 f"{mode.include}({base_mode})-{self.plain}({update_mode})"
        #             ),
        #             content=render(
        #                 f'{mode.include} "{file}{include_suffix}({section})"\n'
        #                 + self.update_template.render(update_mode)
        #             ),
        #             expected=render(self.expected(update_mode)),
        #         )

        # for base_mode, update_mode in itertools.product((MergeMode.Default,), MergeMode):
        #     # Include + Include
        #     mode = MergeMode.Default
        #     base_file = self._make_file(self.base, mode)
        #     base_section = base_file.name
        #     base_file = base_file.parent.name
        #     update_file = self._make_file(self.update, mode)
        #     update_section = update_file.name
        #     update_file = update_file.parent.name
        #     yield Test(
        #         title=self.make_title(
        #             f"{base_mode.include} ({mode}) - {update_mode.include} ({mode})"
        #         ),
        #         file=self.make_file(self.title)
        #         / self.make_file(
        #             f"{base_mode.include}({mode})-{update_mode.include}({mode})"
        #         ),
        #         content=render(
        #             f'{base_mode.include} "{base_file}{include_suffix}({base_section})"\n'
        #             + f'{update_mode.include} "{update_file}{include_suffix}({update_section})"'
        #         ),
        #         expected=render(self.expected(update_mode)),
        #     )

        for base_mode, update_mode in itertools.product(MergeMode, MergeMode):
            # Multi-include
            base_file = self._make_file(self._base, base_mode)
            base_section = base_file.name
            base_file = base_file.parent.name
            update_file = self._make_file(self._update, update_mode)
            update_section = update_file.name
            update_file = update_file.parent.name
            for mode in (MergeMode.Override, MergeMode.Augment, MergeMode.Replace):
                yield Test(
                    title=self.make_title(
                        f"{self._include} ({base_mode} {mode.char} {update_mode})"
                    ),
                    file=self.make_file(self.title) / self.make_file(mode),
                    content=render(
                        f'{base_mode.include} "{base_file}{include_suffix}({base_section})'
                        + mode.char
                        + f'{update_file}{include_suffix}({update_section})"'
                    ),
                    expected=render(self.expected(mode)),
                )

    def _generate_data(self) -> Generator[Test]:
        for type, template in (
            (self._base, self.base_template),
            (self._update, self.update_template),
        ):
            for mode in MergeMode:
                section_name = self._make_section_name(type, mode)
                file = self._make_file(type, mode)
                content = template.render(mode)
                content = self.render_section(content=content, name=section_name) + "\n"
                yield Test(title="", file=file, content=content, expected="")

    def generate_tests(
        self, compiler: XkbCompiler, as_keymap: bool = False
    ) -> TestGroup:
        if as_keymap:

            def render(content: str) -> str:
                return self.render_keymap(content)
        else:

            def render(content: str) -> str:
                return self.render_section(content)

        return TestGroup(self.title, tuple(self._generate_tests(render, compiler)))

    def generate_data(self) -> TestGroup:
        return TestGroup(self.title, tuple(self._generate_data()))

    @classmethod
    def write(
        cls,
        root: Path,
        jinja_env: jinja2.Environment,
        c_file: Path | None,
        xkb: bool,
        compiler: XkbCompiler,
        tests: Sequence[Self],
        debug: bool = False,
    ):
        if c_file is not None and c_file.name:
            c_data = tuple(
                t.generate_tests(as_keymap=True, compiler=compiler) for t in tests
            )
            TestGroup.write_c(
                root=root,
                jinja_env=jinja_env,
                path=c_file,
                compiler=compiler,
                tests=c_data,
            )
        if xkb:
            xkb_data = tuple(t.generate_data() for t in tests)
            TestGroup.write_data(root=root, compiler=compiler, tests=xkb_data)


KeymapTemplate = Template(
    """\
xkb_keymap {
${keycodes}
${types}
${compat}
${symbols}
};"""
)


@dataclass
class KeymapTestTemplate(TestTemplate):
    keycodes: ComponentTestTemplate | None = None
    types: ComponentTestTemplate | None = None
    compat: ComponentTestTemplate | None = None
    symbols: ComponentTestTemplate | None = None

    def __post_init__(self):
        if self.keycodes:
            self.keycodes = dataclasses.replace(self.keycodes, title="")
        if self.types:
            self.types = dataclasses.replace(self.types, title="")
        if self.compat:
            self.compat = dataclasses.replace(self.compat, title="")
        if self.symbols:
            self.symbols = dataclasses.replace(self.symbols, title="")

    def __iter__(self) -> Generator[ComponentTestTemplate]:
        def make_empty_template(component: Component) -> ComponentTestTemplate:
            return ComponentTestTemplate(
                title="",
                component=component,
                base_template=ComponentTemplate(""),
                update_template=ComponentTemplate(""),
                augment="",
                override="",
            )

        yield self.keycodes or make_empty_template(Component.keycodes)
        yield self.types or make_empty_template(Component.types)
        yield self.compat or make_empty_template(Component.compat)
        yield self.symbols or make_empty_template(Component.symbols)

    def _generate_tests(self, compiler: XkbCompiler) -> Generator[Test]:
        groups = tuple(
            c.generate_tests(as_keymap=False, compiler=compiler) for c in self
        )
        keycodes: Test
        types: Test
        compat: Test
        symbols: Test
        for keycodes, types, compat, symbols in zip(*(g.tests for g in groups)):
            title = self.title + keycodes.title
            file = keycodes.file
            content = KeymapTemplate.substitute(
                keycodes=textwrap.indent(keycodes.content, INDENT),
                types=textwrap.indent(types.content, INDENT),
                compat=textwrap.indent(compat.content, INDENT),
                symbols=textwrap.indent(symbols.content, INDENT),
            )
            expected = KeymapTemplate.substitute(
                keycodes=textwrap.indent(keycodes.expected, INDENT),
                types=textwrap.indent(types.expected, INDENT),
                compat=textwrap.indent(compat.expected, INDENT),
                symbols=textwrap.indent(symbols.expected, INDENT),
            )
            yield Test(title=title, file=file, content=content, expected=expected)

    def generate_tests(self, compiler: XkbCompiler) -> TestGroup:
        return TestGroup(self.title, tuple(self._generate_tests(compiler)))

    def _generate_data(self) -> Generator[Test]:
        for component in self:
            yield from component.generate_data().tests

    def generate_data(self) -> TestGroup:
        return TestGroup(self.title, tuple(self._generate_data()))

    @classmethod
    def write(
        cls,
        root: Path,
        jinja_env: jinja2.Environment,
        c_file: Path | None,
        xkb: bool,
        compiler: XkbCompiler,
        tests: Sequence[Self],
        debug: bool = False,
    ):
        if c_file is not None and c_file.name:
            c_data = tuple(t.generate_tests(compiler) for t in tests)
            TestGroup.write_c(
                root=root,
                jinja_env=jinja_env,
                path=c_file,
                compiler=compiler,
                tests=c_data,
            )
        if xkb:
            xkb_data = tuple(t.generate_data() for t in tests)
            TestGroup.write_data(root=root, compiler=compiler, tests=xkb_data)


################################################################################
#
# Key types tests
#
################################################################################

TYPES_TESTS = ComponentTestTemplate(
    "Types",
    component=Component.types,
    base_template=ComponentTemplate(
        """
        ${mode}virtual_modifiers NumLock, LevelThree=none, LevelFive=Mod3;
        ${mode}virtual_modifiers U1, U2, U3;
        // ${mode}virtual_modifiers U4, U5, U6;
        ${mode}virtual_modifiers Z1=none, Z2=none, Z3=none;
        // ${mode}virtual_modifiers Z4=none, Z5=none, Z6=none;
        ${mode}virtual_modifiers M1=0x1000, M2=0x2000, M3=0x3000;
        // ${mode}virtual_modifiers M4=0x4000, M5=0x5000, M6=0x6000;

        ${mode}type "ONE_LEVEL" {
        	modifiers = None;
        	map[None] = Level1;
        	level_name[Level1]= "Any";
        };

        ${mode}type "TWO_LEVEL" {
        	modifiers = Shift;
        	map[Shift] = Level2;
        	level_name[Level1] = "Base";
        	level_name[Level2] = "Shift";
        };

        ${mode}type "ALPHABETIC" {
        	modifiers = Shift+Lock;
        	map[Shift] = Level2;
        	map[Lock] = Level2;
        	level_name[Level1] = "Base";
        	level_name[Level2] = "Caps";
        };

        ${mode}type "KEYPAD" {
        	modifiers = Shift+NumLock;
        	map[None] = Level1;
        	map[Shift] = Level2;
        	map[NumLock] = Level2;
        	map[Shift+NumLock] = Level1;
        	level_name[Level1] = "Base";
        	level_name[Level2] = "Number";
        };

        ${mode}type "FOUR_LEVEL" {
        	modifiers = Shift+LevelThree;
        	map[None] = Level1;
        	map[Shift] = Level2;
        	map[LevelThree] = Level3;
        	map[Shift+LevelThree] = Level4;
        	level_name[Level1] = "Base";
        	level_name[Level2] = "Shift";
        	level_name[Level3] = "Alt Base";
        	level_name[Level4] = "Shift Alt";
        };

        ${mode}type "FOUR_LEVEL_ALPHABETIC" {
        	modifiers = Shift+Lock+LevelThree;
        	map[None] = Level1;
        	map[Shift] = Level2;
        	map[Lock]  = Level2;
        	map[LevelThree] = Level3;
        	map[Shift+LevelThree] = Level4;
        	map[Lock+LevelThree] =  Level4;
        	map[Lock+Shift+LevelThree] =  Level3;
        	level_name[Level1] = "Base";
        	level_name[Level2] = "Shift";
        	level_name[Level3] = "Alt Base";
        	level_name[Level4] = "Shift Alt";
        };

        ${mode}type "FOUR_LEVEL_SEMIALPHABETIC" {
        	modifiers = Shift+Lock+LevelThree;
        	map[None] = Level1;
        	map[Shift] = Level2;
        	map[Lock]  = Level2;
        	map[LevelThree] = Level3;
        	map[Shift+LevelThree] = Level4;
        	map[Lock+LevelThree] =  Level3;
        	map[Lock+Shift+LevelThree] = Level4;
        	preserve[Lock+LevelThree] = Lock;
        	preserve[Lock+Shift+LevelThree] = Lock;
        	level_name[Level1] = "Base";
        	level_name[Level2] = "Shift";
        	level_name[Level3] = "Alt Base";
        	level_name[Level4] = "Shift Alt";
        };

        ${mode}type "XXX" {
        	modifiers = Shift+Lock+LevelThree;
            map[None]       = 1;
            map[Shift]      = 2;
            map[Lock]       = 2;
            map[LevelThree] = 3;
            level_name[1] = "1";
            level_name[2] = "2";
            level_name[3] = "3";
        };
        """
    ),
    update_template=ComponentTemplate(
        """\
        ${mode}virtual_modifiers NumLock=Mod2;    // Changed: now mapped
        ${mode}virtual_modifiers LevelThree=Mod5; // Changed: altered mapping (from 0)
        ${mode}virtual_modifiers LevelFive=Mod3;  // Unhanged: same mapping

        ${mode}virtual_modifiers U7, Z7=0, M7=0x700000; // Changed: new

        ${mode}virtual_modifiers U1;            // Unchanged (unmapped)
        ${mode}virtual_modifiers U2 = none;     // Changed: now mapped
        ${mode}virtual_modifiers U3 = 0x300000; // Changed: now mapped
        // ${mode}virtual_modifiers U4;            // Unchanged (unmapped)
        // ${mode}virtual_modifiers U5 = none;     // Changed: now mapped
        // ${mode}virtual_modifiers U6 = 0x600000; // Changed: now mapped
        ${mode}virtual_modifiers Z1;            // Unchanged (= 0)
        ${mode}virtual_modifiers Z2 = none;     // Unchanged (same mapping)
        ${mode}virtual_modifiers Z3 = 0x310000; // Changed: altered mapping (from 0)
        // ${mode}virtual_modifiers Z4;            // Unchanged (= 0)
        // ${mode}virtual_modifiers Z5 = none;     // Unchanged (same mapping)
        // ${mode}virtual_modifiers Z6 = 0x610000; // Changed: altered mapping (from 0)
        ${mode}virtual_modifiers M1;            // Unchanged (≠ 0)
        ${mode}virtual_modifiers M2 = none;     // Changed: reset
        ${mode}virtual_modifiers M3 = 0x320000; // Changed: altered mapping (from ≠ 0)
        // ${mode}virtual_modifiers M4;            // Unchanged (≠ 0)
        // ${mode}virtual_modifiers M5 = none;     // Changed: reset
        // ${mode}virtual_modifiers M6 = 0x620000; // Changed: altered mapping (from ≠ 0)

        ${mode}type "ONE_LEVEL" {
        	modifiers = None;
        	map[None] = Level1;
        	level_name[Level1]= "New"; // Change name
        };

        ${mode}type "TWO_LEVEL" {
        	modifiers = Shift+M1; // Changed
        	map[Shift] = Level2;
        	map[M1] = Level2; // Changed
        	level_name[Level1] = "Base";
        	level_name[Level2] = "Shift";
        };

        ${mode}type "ALPHABETIC" {
        	modifiers = Lock; // Changed
        	map[None] = Level2; // Changed
        	map[Lock] = Level1; // Changed
        	level_name[Level1] = "Base";
        	level_name[Level2] = "Caps";
        };

        ${mode}type "KEYPAD" {
        	modifiers = Shift+NumLock;
        	map[None] = Level1;
        	map[Shift] = Level2; // Changed
        	map[NumLock] = Level2;
        	map[Shift+NumLock] = Level1;
        	level_name[Level1] = "Base";
        	level_name[Level2] = "Number";
        };

        // Unchanged
        ${mode}type "FOUR_LEVEL" {
        	modifiers = Shift+LevelThree;
        	map[None] = Level1;
        	map[Shift] = Level2;
        	map[LevelThree] = Level3;
        	map[Shift+LevelThree] = Level4;
        	level_name[Level1] = "Base";
        	level_name[Level2] = "Shift";
        	level_name[Level3] = "Alt Base";
        	level_name[Level4] = "Shift Alt";
        };

        ${mode}type "FOUR_LEVEL_ALPHABETIC" {
        	modifiers = Shift+Lock+LevelThree;
        	map[None] = Level1;
        	map[Shift] = Level2;
        	map[Lock]  = Level2;
        	map[LevelThree] = Level3;
        	map[Shift+LevelThree] = Level4;
        	map[Lock+LevelThree] =  Level4;
        	map[Lock+Shift+LevelThree] =  Level3;
        	level_name[Level1] = "Base";
        	level_name[Level2] = "Shift";
        	level_name[Level3] = "Alt Base";
        	level_name[Level4] = "Shift Alt";
        };

        ${mode}type "FOUR_LEVEL_SEMIALPHABETIC" {
        	modifiers = Shift+Lock+LevelThree;
        	map[None] = Level1;
        	map[Shift] = Level2;
        	map[Lock]  = Level2;
        	map[LevelThree] = Level3;
        	map[Shift+LevelThree] = Level4;
        	map[Lock+LevelThree] =  Level3;
        	map[Lock+Shift+LevelThree] = Level4;
        	preserve[Lock+LevelThree] = Lock;
        	preserve[Lock+Shift+LevelThree] = Lock;
        	level_name[Level1] = "Base";
        	level_name[Level2] = "Shift";
        	level_name[Level3] = "Alt Base";
        	level_name[Level4] = "Shift Alt";
        };

        ${mode}type "XXX" {
        	modifiers = Shift+Lock+LevelThree;
            map[None]       = 1;
            map[Shift]      = 2;
            // map[Lock]       = 2;    // Changed
            // map[LevelThree] = 3;
            map[LevelThree] = 1;       // Changed
            map[Shift+LevelThree] = 4; // Changed
            level_name[1] = "A"; // Changed
            level_name[2] = "2";
            level_name[3] = "3";
            level_name[4] = "4"; // Changed
        };

        // New
        ${mode}type "YYY" {
        	modifiers = None;
        	map[None] = Level1;
        	level_name[Level1]= "New";
        };
        """
    ),
    augment="""
        virtual_modifiers NumLock=Mod2,LevelThree=none,LevelFive=Mod3;

        virtual_modifiers U1,U2=0,U3=0x300000;
        // virtual_modifiers U4,U5=0,U6=0x600000;
        virtual_modifiers Z1=0,Z2=0,Z3=0;
        // virtual_modifiers Z4=0,Z5=0,Z6=0;
        virtual_modifiers M1=0x1000,M2=0x2000,M3=0x3000;
        // virtual_modifiers M4=0x4000,M5=0x5000,M6=0x6000;
        virtual_modifiers U7,Z7=0,M7=0x700000;

        type "ONE_LEVEL" {
        	modifiers= none;
            map[none]= 1;
        	level_name[1]= "Any";
        };
        type "TWO_LEVEL" {
        	modifiers= Shift;
        	map[Shift]= 2;
        	level_name[1]= "Base";
        	level_name[2]= "Shift";
        };
        type "ALPHABETIC" {
        	modifiers= Shift+Lock;
        	map[Shift]= 2;
        	map[Lock]= 2;
        	level_name[1]= "Base";
        	level_name[2]= "Caps";
        };
        type "KEYPAD" {
        	modifiers= Shift+NumLock;
        	map[Shift]= 2;
        	map[NumLock]= 2;
        	level_name[1]= "Base";
        	level_name[2]= "Number";
        };
        type "FOUR_LEVEL" {
        	modifiers= Shift+LevelThree;
            map[none]= 1;
        	map[Shift]= 2;
        	map[LevelThree]= 3;
        	map[Shift+LevelThree]= 4;
        	level_name[1]= "Base";
        	level_name[2]= "Shift";
        	level_name[3]= "Alt Base";
        	level_name[4]= "Shift Alt";
        };
        type "FOUR_LEVEL_ALPHABETIC" {
        	modifiers= Shift+Lock+LevelThree;
            map[none]= 1;
        	map[Shift]= 2;
        	map[Lock]= 2;
        	map[LevelThree]= 3;
        	map[Shift+LevelThree]= 4;
        	map[Lock+LevelThree]= 4;
        	map[Shift+Lock+LevelThree]= 3;
        	level_name[1]= "Base";
        	level_name[2]= "Shift";
        	level_name[3]= "Alt Base";
        	level_name[4]= "Shift Alt";
        };
        type "FOUR_LEVEL_SEMIALPHABETIC" {
        	modifiers= Shift+Lock+LevelThree;
        	map[None]= 1;
        	map[Shift]= 2;
        	map[Lock]= 2;
        	map[LevelThree]= 3;
        	map[Shift+LevelThree]= 4;
        	map[Lock+LevelThree]= 3;
        	preserve[Lock+LevelThree]= Lock;
        	map[Shift+Lock+LevelThree]= 4;
        	preserve[Shift+Lock+LevelThree]= Lock;
        	level_name[1]= "Base";
        	level_name[2]= "Shift";
        	level_name[3]= "Alt Base";
        	level_name[4]= "Shift Alt";
        };
        type "XXX" {
        	modifiers = Shift+Lock+LevelThree;
            map[None]       = 1;
            map[Shift]      = 2;
            map[Lock]       = 2;
            map[LevelThree] = 3;
            level_name[1] = "1";
            level_name[2] = "2";
            level_name[3] = "3";
        };
        type "YYY" {
        	modifiers = None;
        	map[None] = Level1;
        	level_name[Level1]= "New";
        };
        """,
    override="""
        virtual_modifiers NumLock=Mod2,LevelThree=Mod5,LevelFive=Mod3;

        virtual_modifiers U1,U2=0,U3=0x300000;
        // virtual_modifiers U4,U5=0,U6=0x600000;
        virtual_modifiers Z1=0,Z2=0,Z3=0x310000;
        // virtual_modifiers Z4=0,Z5=0,Z6=0x610000;
        virtual_modifiers M1=0x1000,M2=0,M3=0x320000;
        // virtual_modifiers M4=0x4000,M5=0,M6=0x620000;
        virtual_modifiers U7,Z7=0,M7=0x700000;

        type "ONE_LEVEL" {
        	modifiers= none;
        	map[None] = Level1;
        	level_name[1]= "New";
        };
        type "TWO_LEVEL" {
        	modifiers= Shift+M1;
        	map[Shift]= 2;
        	map[M1]= 2;
        	level_name[1]= "Base";
        	level_name[2]= "Shift";
        };
        type "ALPHABETIC" {
        	modifiers= Lock;
        	map[none]= 2;
        	map[Lock]= 1;
        	level_name[1]= "Base";
        	level_name[2]= "Caps";
        };
        type "KEYPAD" {
        	modifiers= Shift+NumLock;
        	map[Shift] = Level2;
        	map[NumLock]= 2;
        	level_name[1]= "Base";
        	level_name[2]= "Number";
        };
        type "FOUR_LEVEL" {
        	modifiers= Shift+LevelThree;
        	map[None]= 1;
        	map[Shift]= 2;
        	map[LevelThree]= 3;
        	map[Shift+LevelThree] = Level4;
        	level_name[1]= "Base";
        	level_name[2]= "Shift";
        	level_name[3]= "Alt Base";
        	level_name[4]= "Shift Alt";
        };
        type "FOUR_LEVEL_ALPHABETIC" {
        	modifiers= Shift+Lock+LevelThree;
        	map[Shift]= 2;
        	map[Lock]= 2;
        	map[LevelThree]= 3;
        	map[Shift+LevelThree]= 4;
        	map[Lock+LevelThree]= 4;
        	map[Shift+Lock+LevelThree]= 3;
        	level_name[1]= "Base";
        	level_name[2]= "Shift";
        	level_name[3]= "Alt Base";
        	level_name[4]= "Shift Alt";
        };
        type "FOUR_LEVEL_SEMIALPHABETIC" {
        	modifiers= Shift+Lock+LevelThree;
        	map[None]= 1;
        	map[Shift]= 2;
        	map[Lock]= 2;
        	map[LevelThree]= 3;
        	map[Shift+LevelThree]= 4;
        	map[Lock+LevelThree]= 3;
        	preserve[Lock+LevelThree]= Lock;
        	map[Shift+Lock+LevelThree]= 4;
        	preserve[Shift+Lock+LevelThree]= Lock;
        	level_name[1]= "Base";
        	level_name[2]= "Shift";
        	level_name[3]= "Alt Base";
        	level_name[4]= "Shift Alt";
        };
        type "XXX" {
        	modifiers = Shift+Lock+LevelThree;
            map[None]       = 1;
            map[Shift]      = 2;
            // map[LevelThree] = 3;
            map[LevelThree] = 1;
            map[Shift+LevelThree] = 4;
            level_name[1] = "A";
            level_name[2] = "2";
            level_name[3] = "3";
            level_name[4] = "4";
        };
        type "YYY" {
        	modifiers = None;
        	map[None] = Level1;
        	level_name[Level1]= "New";
        };
        """,
)


################################################################################
#
# Compat tests
#
################################################################################

COMPAT_TESTS = ComponentTestTemplate(
    "Compat",
    component=Component.compat,
    base_template=ComponentTemplate(
        """
        ${mode}virtual_modifiers NumLock;

        interpret.repeat= False;
        setMods.clearLocks= True;
        latchMods.clearLocks= True;
        latchMods.latchToLock= True;

        ${mode}interpret Any + Any {
        	action= SetMods(modifiers=modMapMods);
        };

        ${mode}interpret Caps_Lock {
        	action = LockMods(modifiers = Lock);
        };

        ${mode}indicator "Caps Lock" {
        	!allowExplicit;
        	whichModState= Locked;
        	modifiers= Lock;
        };

        ${mode}indicator "Num Lock" {
        	!allowExplicit;
        	whichModState= Locked;
        	modifiers= NumLock;
        };
        """
    ),
    update_template=ComponentTemplate(
        """
        ${mode}virtual_modifiers NumLock;

        ${mode}interpret.repeat= False;
        ${mode}setMods.clearLocks= False; // Changed

        // Unchanged
        ${mode}interpret Any + Any {
        	action= SetMods(modifiers=modMapMods);
        };

        // Changed
        ${mode}interpret Caps_Lock {
        	action = LockMods(modifiers = NumLock);
        };

        // Unchanged
        ${mode}indicator "Caps Lock" {
        	!allowExplicit;
        	whichModState= Locked;
        	modifiers= Lock;
        };

        // Changed
        ${mode}indicator "Num Lock" {
        	!allowExplicit;
        	whichModState= Base;
        	modifiers= Lock;
        };

        // New
        ${mode}indicator "Kana" {
        	!allowExplicit;
        	whichModState= Locked;
        	modifiers= Control;
        };
        """
    ),
    augment="""
        virtual_modifiers NumLock;

        interpret.useModMapMods= AnyLevel;
        interpret.repeat= False;
        interpret Caps_Lock+AnyOfOrNone(all) {
        	action= LockMods(modifiers=Lock);
        };
        interpret Any+AnyOf(all) {
        	action= SetMods(modifiers=modMapMods,clearLocks);
        };
        indicator "Caps Lock" {
        	whichModState= locked;
        	modifiers= Lock;
        };
        indicator "Num Lock" {
        	whichModState= locked;
        	modifiers= NumLock;
        };
        indicator "Kana" {
        	whichModState= locked;
        	modifiers= Control;
	    };
        """,
    override="""
        virtual_modifiers NumLock;

        interpret.repeat= False;
        setMods.clearLocks= False;
        interpret Caps_Lock+AnyOfOrNone(all) {
        	action= LockMods(modifiers=NumLock);
        };
        interpret Any+AnyOf(all) {
        	action= SetMods(modifiers=modMapMods);
        };
        indicator "Caps Lock" {
        	whichModState= locked;
        	modifiers= Lock;
        };
        indicator "Kana" {
        	whichModState= locked;
        	modifiers= Control;
        };
        indicator "Num Lock" {
        	whichModState= base;
        	modifiers= Lock;
        };
        """,
)


################################################################################
#
# Symbols tests
#
################################################################################


LINUX_EVENT_CODES = (
    None,
    "ESC",
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
    "0",
    "MINUS",
    "EQUAL",
    "BACKSPACE",
    "TAB",
    "Q",
    "W",
    "E",
    "R",
    "T",
    "Y",
    "U",
    "I",
    "O",
    "P",
    "LEFTBRACE",
    "RIGHTBRACE",
    "ENTER",
    "LEFTCTRL",
    "A",
    "S",
    "D",
    "F",
    "G",
    "H",
    "J",
    "K",
    "L",
    "SEMICOLON",
    "APOSTROPHE",
    "GRAVE",
    "LEFTSHIFT",
    "BACKSLASH",
    "Z",
    "X",
    "C",
    "V",
    "B",
    "N",
    "M",
    "COMMA",
    "DOT",
    "SLASH",
    "RIGHTSHIFT",
    "KPASTERISK",
    "LEFTALT",
    "SPACE",
    "CAPSLOCK",
    "F1",
    "F2",
    "F3",
    "F4",
    "F5",
    "F6",
    "F7",
    "F8",
    "F9",
    "F10",
    "NUMLOCK",
    "SCROLLLOCK",
    "KP7",
    "KP8",
    "KP9",
    "KPMINUS",
    "KP4",
    "KP5",
    "KP6",
    "KPPLUS",
    "KP1",
    "KP2",
    "KP3",
    "KP0",
    "KPDOT",
    None,
    "ZENKAKUHANKAKU",
    "102ND",
    "F11",
    "F12",
    "RO",
    "KATAKANA",
    "HIRAGANA",
    "HENKAN",
    "KATAKANAHIRAGANA",
    "MUHENKAN",
    "KPJPCOMMA",
    "KPENTER",
    "RIGHTCTRL",
    "KPSLASH",
    "SYSRQ",
    "RIGHTALT",
    "LINEFEED",
    "HOME",
    "UP",
    "PAGEUP",
    "LEFT",
    "RIGHT",
    "END",
    "DOWN",
    "PAGEDOWN",
    "INSERT",
    "DELETE",
    "MACRO",
    "MUTE",
    "VOLUMEDOWN",
    "VOLUMEUP",
    "POWER",
    "KPEQUAL",
    "KPPLUSMINUS",
    "PAUSE",
    "SCALE",
    "KPCOMMA",
    "HANGEUL",
    "HANJA",
    "YEN",
    "LEFTMETA",
    "RIGHTMETA",
    "COMPOSE",
    "STOP",
    "AGAIN",
    "PROPS",
    "UNDO",
    "FRONT",
    "COPY",
    "OPEN",
    "PASTE",
    "FIND",
    "CUT",
    "HELP",
    "MENU",
    "CALC",
    "SETUP",
    "SLEEP",
    "WAKEUP",
    "FILE",
    "SENDFILE",
    "DELETEFILE",
    "XFER",
    "PROG1",
    "PROG2",
    "WWW",
    "MSDOS",
    "COFFEE",
    "DIRECTION",
    "CYCLEWINDOWS",
    "MAIL",
    "BOOKMARKS",
    "COMPUTER",
    "BACK",
    "FORWARD",
    "CLOSECD",
    "EJECTCD",
    "EJECTCLOSECD",
    "NEXTSONG",
    "PLAYPAUSE",
    "PREVIOUSSONG",
    "STOPCD",
    "RECORD",
    "REWIND",
    "PHONE",
    "ISO",
    "CONFIG",
    "HOMEPAGE",
    "REFRESH",
    "EXIT",
    "MOVE",
    "EDIT",
    "SCROLLUP",
    "SCROLLDOWN",
    "KPLEFTPAREN",
    "KPRIGHTPAREN",
    "NEW",
    "REDO",
    "F13",
    "F14",
    "F15",
    "F16",
    "F17",
    "F18",
    "F19",
    "F20",
    "F21",
    "F22",
    "F23",
    "F24",
    None,
    None,
    None,
    None,
    None,
    "PLAYCD",
    "PAUSECD",
    "PROG3",
    "PROG4",
    "DASHBOARD",
    "SUSPEND",
    "CLOSE",
    "PLAY",
    "FASTFORWARD",
    "BASSBOOST",
    "PRINT",
    "HP",
    "CAMERA",
    "SOUND",
    "QUESTION",
)

Comment = NewType("Comment", str)


@dataclass(frozen=True)
class KeyCode:
    _evdev: str
    _xkb: str

    @property
    def c(self) -> str:
        return f"KEY_{self._evdev}"

    @property
    def xkb(self) -> str:
        return f"<{self._xkb}>"


class Keysym(str):
    @property
    def c(self) -> str:
        return f"XKB_KEY_{self}"

    @classmethod
    def parse(cls, raw: str | None) -> Self:
        if not raw:
            return cls("NoSymbol")
        else:
            return cls(raw)


NoSymbol = Keysym("NoSymbol")


class Modifier(IntFlag):
    NoModifier = 0
    Shift = 1 << 0
    Lock = 1 << 1
    Control = 1 << 2
    Mod1 = 1 << 3
    Mod2 = 1 << 4
    Mod3 = 1 << 5
    Mod4 = 1 << 6
    Mod5 = 1 << 7
    LevelThree = Mod5

    def __iter__(self) -> Iterator[Self]:
        for m in self.__class__:
            if m & self:
                yield m

    def __str__(self) -> str:
        return "+".join(m.name for m in self)


class Action(metaclass=ABCMeta):
    @abstractmethod
    def __bool__(self) -> bool: ...

    @classmethod
    def parse(cls, raw: Any) -> Self:
        if raw is None:
            return GroupAction.parse(None)
        elif isinstance(raw, Modifier):
            return ModAction.parse(raw)
        elif isinstance(raw, int):
            return GroupAction.parse(raw)
        else:
            raise ValueError(raw)

    @abstractmethod
    def action_to_keysym(self, index: int, level: int) -> Keysym: ...


@dataclass
class GroupAction(Action):
    """
    SetGroup or NoAction
    """

    group: int
    keysyms: ClassVar[dict[tuple[int, int, int], Keysym]] = {
        (2, 0, 0): Keysym("a"),
        (2, 0, 1): Keysym("A"),
        (2, 1, 0): Keysym("b"),
        (2, 1, 1): Keysym("B"),
        (3, 0, 0): Keysym("Greek_alpha"),
        (3, 0, 1): Keysym("Greek_ALPHA"),
        (3, 1, 0): Keysym("Greek_beta"),
        (3, 1, 1): Keysym("Greek_BETA"),
    }

    def __str__(self) -> str:
        if self.group > 0:
            return f"SetGroup(group={self.group})"
        else:
            return "NoAction()"

    def __bool__(self) -> bool:
        return bool(self.group)

    @classmethod
    def parse(cls, raw: int | None) -> Self:
        if not raw:
            return cls(0)
        else:
            return cls(raw)

    def action_to_keysym(self, index: int, level: int) -> Keysym:
        if not self.group:
            return NoSymbol
        else:
            if (
                keysym := self.keysyms.get((self.group % 4, index % 2, level % 2))
            ) is None:
                raise ValueError((self, index, level))
            return keysym


@dataclass
class ModAction(Action):
    """
    SetMod or NoAction
    """

    mods: Modifier
    keysyms: ClassVar[dict[tuple[Modifier, int, int], Keysym]] = {
        (Modifier.Control, 0, 0): Keysym("x"),
        (Modifier.Control, 0, 1): Keysym("X"),
        (Modifier.Control, 1, 0): Keysym("y"),
        (Modifier.Control, 1, 1): Keysym("Y"),
        (Modifier.Mod5, 0, 0): Keysym("Greek_xi"),
        (Modifier.Mod5, 0, 1): Keysym("Greek_XI"),
        (Modifier.Mod5, 1, 0): Keysym("Greek_upsilon"),
        (Modifier.Mod5, 1, 1): Keysym("Greek_UPSILON"),
    }

    def __str__(self) -> str:
        if self.mods is Modifier.NoModifier:
            return "NoAction()"
        else:
            return f"SetMods(mods={self.mods})"

    def __bool__(self) -> bool:
        return self.mods is Modifier.NoModifier

    @classmethod
    def parse(cls, raw: Modifier | None) -> Self:
        if not raw:
            return cls(Modifier.NoModifier)
        else:
            return cls(raw)

    def action_to_keysym(self, index: int, level: int) -> Keysym:
        if self.mods is Modifier.NoModifier:
            return NoSymbol
        else:
            if (keysym := self.keysyms.get((self.mods, index % 2, level % 2))) is None:
                raise ValueError((self, index, level))
            return keysym


@dataclass
class Level:
    keysyms: tuple[Keysym, ...]
    actions: tuple[Action, ...]

    @staticmethod
    def _c(default: str, xs: Iterable[Any]) -> str:
        match len(xs):
            case 0:
                return default
            case 1:
                return xs[0].c
            case _:
                return ", ".join(map(lambda x: x.c, xs))

    @staticmethod
    def _xkb(default: str, xs: Iterable[Any]) -> str:
        match len(xs):
            case 0:
                return default
            case 1:
                return str(xs[0])
            case _:
                return "{" + ", ".join(map(str, xs)) + "}"

    @classmethod
    def has_empty_symbols(cls, keysyms: tuple[Keysym, ...]) -> bool:
        return all(ks == NoSymbol for ks in keysyms)

    @property
    def empty_symbols(self) -> bool:
        return self.has_empty_symbols(self.keysyms)

    @property
    def keysyms_c(self) -> str:
        if not self.keysyms and self.actions:
            return self._c(
                NoSymbol.c, tuple(itertools.repeat(NoSymbol, len(self.actions)))
            )
        return self._c(NoSymbol.c, self.keysyms)

    @property
    def keysyms_xkb(self) -> str:
        return self._xkb(NoSymbol, self.keysyms)

    @classmethod
    def has_empty_actions(cls, actions: tuple[Action, ...]) -> bool:
        return not any(actions)

    @property
    def empty_actions(self) -> bool:
        return self.has_empty_actions(self.actions)

    @property
    def actions_xkb(self) -> str:
        return self._xkb("NoAction()", self.actions)

    @classmethod
    def Keysyms(cls, *keysyms: str | None) -> Self:
        return cls.Mix(keysyms, ())

    @classmethod
    def Actions(cls, *actions: int | Modifier | None) -> Self:
        return cls.Mix((), actions)

    @classmethod
    def Mix(
        cls, keysyms: tuple[str | None, ...], actions: tuple[int | Modifier | None,]
    ) -> Self:
        return cls(tuple(map(Keysym.parse, keysyms)), tuple(map(Action.parse, actions)))

    def add_keysyms(self, keep_actions: bool, level: int) -> Self:
        return self.__class__(
            keysyms=tuple(
                a.action_to_keysym(index=k, level=level)
                for k, a in enumerate(self.actions)
            ),
            actions=self.actions if keep_actions else (),
        )

    @property
    def target_group(self) -> int:
        for a in self.actions:
            if isinstance(a, GroupAction) and a.group > 1:
                return a.group
        else:
            return 0

    @property
    def target_level(self) -> int:
        for a in self.actions:
            if isinstance(a, ModAction) and a.mods:
                match a.mods:
                    case Modifier.LevelThree:
                        return 2
                    case _:
                        return 0
        else:
            return 0


@dataclass
class KeyEntry:
    levels: tuple[Level, ...]

    def __init__(self, *levels: Level):
        self.levels = levels

    @property
    def xkb(self) -> Iterator[str]:
        if not self.levels:
            yield ""
            return
        keysyms = tuple(l.keysyms for l in self.levels)
        has_keysyms = any(not Level.has_empty_symbols(s) for s in keysyms)
        no_keysyms = all(not s for s in keysyms)
        actions = tuple(l.actions for l in self.levels)
        has_actions = any(not Level.has_empty_actions(a) for a in actions)
        if has_keysyms or (not no_keysyms and not has_actions):
            yield "["
            yield ", ".join(l.keysyms_xkb for l in self.levels)
            yield "]"
        if has_actions or no_keysyms:
            if has_keysyms:
                yield ", "
            yield "["
            yield ", ".join(l.actions_xkb for l in self.levels)
            yield "]"

    def add_keysyms(self, keep_actions: bool) -> Self:
        return self.__class__(
            *(
                l.add_keysyms(keep_actions=keep_actions, level=k)
                for k, l in enumerate(self.levels)
            )
        )


class TestType(IntFlag):
    KeysymsOnly = auto()
    ActionsOnly = auto()
    KeysymsAndActions = auto()
    All = ActionsOnly | KeysymsOnly | KeysymsAndActions


@dataclass
class TestId:
    type: TestType
    base: int = 0
    _last_base: ClassVar[int] = 0
    _max_type: ClassVar[int] = TestType.KeysymsAndActions >> 1
    _forbidden_keys: tuple[str, ...] = ("LEFTSHIFT", "RIGHTALT")

    def __post_init__(self):
        if self.base == -1:
            self.base = self.__class__._last_base
        elif self.base < 0:
            raise ValueError(self.base)
        elif self.base == 0:
            self.__class__._last_base += 1
            self.base = self.__class__._last_base
            while not self.check_linux_keys(self._base_id):
                self.__class__._last_base += 1
                self.base = self.__class__._last_base
        else:
            self.__class__._last_base = max(self.__class__._last_base, self.base)
        assert self.linux_key, self
        # print(self.base, self.id, self.xkb_key)

    def __hash__(self) -> int:
        return self.id

    @property
    def _base_id(self) -> int:
        return (self.base - 1) * (self._max_type + 1) + 1

    @property
    def id(self) -> int:
        # return (self.base << 2) | (self.type >> 1)
        return self._base_id + (self.type >> 1)

    def with_type(self, type: TestType) -> Self:
        return dataclasses.replace(self, type=type)

    @property
    def xkb_key(self) -> str:
        return f"T{self.id:0>3}"

    @classmethod
    def check_linux_key(cls, idx: int) -> bool:
        if idx >= len(LINUX_EVENT_CODES):
            raise ValueError("Not enough Linux keys available!")
        key: str | None = LINUX_EVENT_CODES[idx]
        return bool(key) and key not in cls._forbidden_keys

    @classmethod
    def check_linux_keys(cls, base_id) -> bool:
        return all(
            cls.check_linux_key(base_id + k) for k in range(0, cls._max_type + 1)
        )

    @property
    def linux_key(self) -> str:
        if self.id >= len(LINUX_EVENT_CODES):
            raise ValueError("Not enough Linux keys available!")
        return LINUX_EVENT_CODES[self.id]

    @property
    def key(self) -> KeyCode:
        return KeyCode(self.linux_key, self.xkb_key)


@unique
class Implementation(Flag):
    x11 = auto()
    xkbcommon = auto()
    all = x11 | xkbcommon


@dataclass
class TestEntry:
    id: TestId
    key: KeyCode = dataclasses.field(init=False)
    base: KeyEntry
    update: KeyEntry
    augment: KeyEntry
    override: KeyEntry
    replace: KeyEntry
    types: TestType
    implementations: Implementation

    group_keysyms: ClassVar[tuple[tuple[str, str], ...]] = (
        ("Ukrainian_i", "Ukrainian_I", "Ukrainian_yi", "Ukrainian_YI"),
        ("ch", "Ch", "c_h", "C_h"),
    )

    def __init__(
        self,
        id: TestId | None,
        base: KeyEntry,
        update: KeyEntry,
        augment: KeyEntry,
        override: KeyEntry,
        types: TestType = TestType.All,
        replace: KeyEntry | None = None,
        implementations: Implementation = Implementation.all,
    ):
        self.id = TestId(0, 0) if id is None else id
        self.key = id.key
        self.base = base
        self.update = update
        self.augment = augment
        self.override = override
        self.types = types
        self.replace = self.update if replace is None else replace
        self.implementations = implementations

    @property
    def default(self) -> KeyEntry:
        return self.override

    def add_keysyms(self, keep_actions: bool) -> Self:
        types = TestType.KeysymsAndActions if keep_actions else TestType.KeysymsOnly
        return dataclasses.replace(
            self,
            id=self.id.with_type(types),
            base=self.base.add_keysyms(keep_actions=keep_actions),
            update=self.update.add_keysyms(keep_actions=keep_actions),
            augment=self.augment.add_keysyms(keep_actions=keep_actions),
            override=self.override.add_keysyms(keep_actions=keep_actions),
            replace=self.replace.add_keysyms(keep_actions=keep_actions),
            types=types,
        )

    @classmethod
    def alt_keysym(cls, group: int, level: int) -> Keysym:
        return Keysym(cls.group_keysyms[group % 2][level % 4])

    @classmethod
    def alt_keysyms(cls, group: int) -> Iterator[Keysym]:
        for keysym in cls.group_keysyms[group % 2]:
            yield Keysym(keysym)


@dataclass
class SymbolsTestGroup:
    name: str
    tests: tuple[TestEntry | Comment, ...]

    def _with_implementation(
        self, implementation: Implementation
    ) -> Iterable[TestEntry | Comment]:
        pending_comment: Comment | None = None
        for t in self.tests:
            if not isinstance(t, TestEntry):
                pending_comment = t
            elif t.implementations & implementation:
                if pending_comment is not None:
                    yield pending_comment
                    pending_comment = None
                yield t

    def with_implementation(self, implementation: Implementation) -> Self:
        return dataclasses.replace(
            self, tests=tuple(self._with_implementation(implementation))
        )

    @staticmethod
    def _add_keysyms(entry: TestEntry | Comment) -> Iterable[TestEntry | Comment]:
        if isinstance(entry, TestEntry) and entry.id.type is TestType.ActionsOnly:
            if entry.types & TestType.KeysymsOnly:
                yield entry.add_keysyms(keep_actions=False)
            yield entry
            if entry.types & TestType.KeysymsAndActions:
                yield entry.add_keysyms(keep_actions=True)
        else:
            yield entry

    def add_keysyms(self, name: str = "") -> Self:
        return dataclasses.replace(
            self,
            name=name or self.name,
            tests=tuple(t for ts in self.tests for t in self._add_keysyms(ts)),
        )

    def __add__(self, other: Any) -> Self:
        if isinstance(other, tuple):
            return dataclasses.replace(self, tests=self.tests + other)
        elif isinstance(other, self.__class__):
            return dataclasses.replace(self, tests=self.tests + other.tests)
        else:
            return NotImplemented

    def get_keycodes(self, base: ComponentTestTemplate) -> ComponentTestTemplate:
        ids: set[TestId] = set(t.id for t in self.tests if isinstance(t, TestEntry))

        base_template = Template(
            "\n".join(f"${{mode}}<{id.xkb_key}> = {id.id + 8};" for id in ids) + "\n"
        )
        keycodes = base_template.substitute(mode="")

        return dataclasses.replace(
            base,
            base_template=base_template + base.base_template,
            update_template=base.update_template,
            augment=keycodes + base.augment,
            override=keycodes + base.override,
            replace=keycodes + base.replace,
        )

    def _get_symbols(self, type: str, debug: bool) -> Generator[str]:
        for entry in self.tests:
            if isinstance(entry, str):
                if type == "update":
                    yield f"\n////// {entry} //////"
            else:
                data: KeyEntry = getattr(entry, type)
                if debug:
                    yield f"// base:     {''.join(entry.base.xkb) or '(empty)'}"
                    yield "// key to merge / replace result"
                yield f"${{mode}}key {entry.key.xkb} {{ {''.join(data.xkb)} }};"
                if debug:
                    yield f"// override: {''.join(entry.override.xkb) or '(empty)'}"
                    yield f"// augment:  {''.join(entry.augment.xkb) or '(empty)'}"

    def get_symbols(self, debug: bool = False) -> ComponentTestTemplate:
        base_template = ComponentTemplate("\n".join(self._get_symbols("base", False)))
        update_template = ComponentTemplate(
            "\n".join(self._get_symbols("update", debug))
        )
        augment = ComponentTemplate(
            "\n".join(self._get_symbols("augment", False))
        ).substitute(mode="")
        override = ComponentTemplate(
            "\n".join(self._get_symbols("override", False))
        ).substitute(mode="")
        replace = ComponentTemplate(
            "\n".join(self._get_symbols("replace", False))
        ).substitute(mode="")
        return ComponentTestTemplate(
            title=self.name,
            component=Component.symbols,
            base_template=base_template,
            update_template=update_template,
            augment=augment,
            override=override,
            replace=replace,
        )


SYMBOLS_TESTS_BOTH = SymbolsTestGroup(
    "",
    (
        Comment("Trivial cases"),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(),
            update=KeyEntry(),
            augment=KeyEntry(),
            override=KeyEntry(),
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(),
            update=KeyEntry(Level.Actions(3)),
            augment=KeyEntry(Level.Actions(3)),
            override=KeyEntry(Level.Actions(3)),
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2)),
            update=KeyEntry(),
            augment=KeyEntry(Level.Actions(2)),
            override=KeyEntry(Level.Actions(2)),
        ),
        Comment("Same key"),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2)),
            update=KeyEntry(Level.Actions(2)),
            augment=KeyEntry(Level.Actions(2)),
            override=KeyEntry(Level.Actions(2)),
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2), Level.Actions(2, Modifier.Control)),
            update=KeyEntry(Level.Actions(2), Level.Actions(2, Modifier.Control)),
            augment=KeyEntry(Level.Actions(2), Level.Actions(2, Modifier.Control)),
            override=KeyEntry(Level.Actions(2), Level.Actions(2, Modifier.Control)),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Mismatch levels count"),
        (
            TEST_BOTH_Q := TestEntry(
                TestId(TestType.ActionsOnly),
                KeyEntry(Level.Actions(None), Level.Actions(2)),
                update=KeyEntry(
                    Level.Actions(3), Level.Actions(None), Level.Actions(None)
                ),
                augment=KeyEntry(
                    Level.Actions(3), Level.Actions(2), Level.Actions(None)
                ),
                override=KeyEntry(
                    Level.Actions(3), Level.Actions(2), Level.Actions(None)
                ),
                # X11 and xkbcommon handle keysyms-only case differently
                types=TestType.ActionsOnly | TestType.KeysymsAndActions,
            )
        ),
        # Trailing NoSymbols are discarded in xkbcomp
        dataclasses.replace(
            TEST_BOTH_Q,
            augment=KeyEntry(Level.Actions(3), Level.Actions(2)),
            override=KeyEntry(Level.Actions(3), Level.Actions(2)),
            replace=KeyEntry(Level.Actions(3)),
            implementations=Implementation.x11,
        ).add_keysyms(keep_actions=False),
        dataclasses.replace(
            TEST_BOTH_Q, implementations=Implementation.xkbcommon
        ).add_keysyms(keep_actions=False),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None), Level.Actions(2)),
            update=KeyEntry(Level.Actions(3), Level.Actions(None), Level.Actions(None)),
            augment=KeyEntry(Level.Actions(3), Level.Actions(2), Level.Actions(None)),
            override=KeyEntry(Level.Actions(3), Level.Actions(2), Level.Actions(None)),
            # X11 and xkbcommon handle keysyms-only case differently
            types=TestType.ActionsOnly | TestType.KeysymsAndActions,
        ),
        (
            TEST_BOTH_W := TestEntry(
                TestId(TestType.ActionsOnly),
                KeyEntry(Level.Actions(None), Level.Actions(2), Level.Actions(None)),
                update=KeyEntry(Level.Actions(3), Level.Actions(None)),
                augment=KeyEntry(
                    Level.Actions(3), Level.Actions(2), Level.Actions(None)
                ),
                override=KeyEntry(
                    Level.Actions(3), Level.Actions(2), Level.Actions(None)
                ),
                # X11 and xkbcommon handle keysyms-only case differently
                types=TestType.ActionsOnly | TestType.KeysymsAndActions,
            )
        ),
        dataclasses.replace(
            TEST_BOTH_W,
            augment=KeyEntry(Level.Actions(3), Level.Actions(2)),
            override=KeyEntry(Level.Actions(3), Level.Actions(2)),
            replace=KeyEntry(Level.Actions(3)),
            implementations=Implementation.x11,
        ).add_keysyms(keep_actions=False),
        dataclasses.replace(
            TEST_BOTH_W, implementations=Implementation.xkbcommon
        ).add_keysyms(keep_actions=False),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2), Level.Actions(2)),
            update=KeyEntry(Level.Actions(3), Level.Actions(3), Level.Actions(3)),
            augment=KeyEntry(Level.Actions(2), Level.Actions(2), Level.Actions(3)),
            override=KeyEntry(Level.Actions(3), Level.Actions(3), Level.Actions(3)),
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2), Level.Actions(2), Level.Actions(2)),
            update=KeyEntry(Level.Actions(3), Level.Actions(3)),
            augment=KeyEntry(Level.Actions(2), Level.Actions(2), Level.Actions(2)),
            override=KeyEntry(Level.Actions(3), Level.Actions(3), Level.Actions(2)),
        ),
        Comment("Single keysyms -> single keysyms"),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None), Level.Actions(None)),
            update=KeyEntry(Level.Actions(None), Level.Actions(None)),
            augment=KeyEntry(Level.Actions(None), Level.Actions(None)),
            override=KeyEntry(Level.Actions(None), Level.Actions(None)),
        ),
        (
            TEST_BOTH_Y := TestEntry(
                TestId(TestType.ActionsOnly),
                KeyEntry(Level.Actions(None), Level.Actions(None)),
                update=KeyEntry(Level.Actions(3), Level.Actions(None)),
                augment=KeyEntry(Level.Actions(3), Level.Actions(None)),
                override=KeyEntry(Level.Actions(3), Level.Actions(None)),
                # X11 and xkbcommon handle keysyms-only case differently
                types=TestType.ActionsOnly | TestType.KeysymsAndActions,
            )
        ),
        dataclasses.replace(
            TEST_BOTH_Y,
            augment=KeyEntry(Level.Actions(3)),
            override=KeyEntry(Level.Actions(3)),
            replace=KeyEntry(Level.Actions(3)),
            implementations=Implementation.x11,
        ).add_keysyms(keep_actions=False),
        dataclasses.replace(
            TEST_BOTH_Y, implementations=Implementation.xkbcommon
        ).add_keysyms(keep_actions=False),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None), Level.Actions(None)),
            update=KeyEntry(Level.Actions(None), Level.Actions(3)),
            augment=KeyEntry(Level.Actions(None), Level.Actions(3)),
            override=KeyEntry(Level.Actions(None), Level.Actions(3)),
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None), Level.Actions(None)),
            update=KeyEntry(Level.Actions(3), Level.Actions(3)),
            augment=KeyEntry(Level.Actions(3), Level.Actions(3)),
            override=KeyEntry(Level.Actions(3), Level.Actions(3)),
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2), Level.Actions(2)),
            update=KeyEntry(Level.Actions(None), Level.Actions(None)),
            augment=KeyEntry(Level.Actions(2), Level.Actions(2)),
            override=KeyEntry(Level.Actions(2), Level.Actions(2)),
        ),
        (
            TEST_BOTH_P := TestEntry(
                TestId(TestType.ActionsOnly),
                KeyEntry(Level.Actions(2), Level.Actions(2)),
                update=KeyEntry(Level.Actions(3), Level.Actions(None)),
                augment=KeyEntry(Level.Actions(2), Level.Actions(2)),
                override=KeyEntry(Level.Actions(3), Level.Actions(2)),
                # X11 and xkbcommon handle keysyms-only case differently
                types=TestType.ActionsOnly | TestType.KeysymsAndActions,
            )
        ),
        dataclasses.replace(
            TEST_BOTH_P,
            augment=KeyEntry(Level.Actions(2), Level.Actions(2)),
            override=KeyEntry(Level.Actions(3), Level.Actions(2)),
            replace=KeyEntry(Level.Actions(3)),
            implementations=Implementation.x11,
        ).add_keysyms(keep_actions=False),
        dataclasses.replace(
            TEST_BOTH_P, implementations=Implementation.xkbcommon
        ).add_keysyms(keep_actions=False),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2), Level.Actions(2)),
            update=KeyEntry(Level.Actions(None), Level.Actions(3)),
            augment=KeyEntry(Level.Actions(2), Level.Actions(2)),
            override=KeyEntry(Level.Actions(2), Level.Actions(3)),
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2), Level.Actions(2)),
            update=KeyEntry(Level.Actions(3), Level.Actions(3)),
            augment=KeyEntry(Level.Actions(2), Level.Actions(2)),
            override=KeyEntry(Level.Actions(3), Level.Actions(3)),
        ),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(Level.Keysyms("a"), Level.Actions(2)),
            update=KeyEntry(Level.Actions(3), Level.Keysyms("X")),
            augment=KeyEntry(Level.Mix(("a"), (3,)), Level.Mix(("X",), (2,))),
            override=KeyEntry(Level.Mix(("a"), (3,)), Level.Mix(("X",), (2,))),
        ),
        Comment("Single keysyms -> multiple keysyms"),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None), Level.Actions(None)),
            update=KeyEntry(Level.Actions(3, None), Level.Actions(None)),
            augment=KeyEntry(Level.Actions(3, None), Level.Actions(None)),
            override=KeyEntry(Level.Actions(3, None), Level.Actions(None)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None), Level.Actions(None)),
            update=KeyEntry(Level.Actions(3, None), Level.Actions(None, None)),
            augment=KeyEntry(Level.Actions(3, None), Level.Actions(None)),
            override=KeyEntry(Level.Actions(3, None), Level.Actions(None)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None), Level.Actions(None)),
            update=KeyEntry(Level.Actions(None), Level.Actions(3, None)),
            augment=KeyEntry(Level.Actions(None), Level.Actions(3, None)),
            override=KeyEntry(Level.Actions(None), Level.Actions(3, None)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None), Level.Actions(None)),
            update=KeyEntry(Level.Actions(None, None), Level.Actions(3, None)),
            augment=KeyEntry(Level.Actions(None), Level.Actions(3, None)),
            override=KeyEntry(Level.Actions(None), Level.Actions(3, None)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None), Level.Actions(None)),
            update=KeyEntry(Level.Actions(3, None), Level.Actions(3, None)),
            augment=KeyEntry(Level.Actions(3, None), Level.Actions(3, None)),
            override=KeyEntry(Level.Actions(3, None), Level.Actions(3, None)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None), Level.Actions(None)),
            update=KeyEntry(
                Level.Actions(3, Modifier.LevelThree),
                Level.Actions(3, Modifier.LevelThree),
            ),
            augment=KeyEntry(
                Level.Actions(3, Modifier.LevelThree),
                Level.Actions(3, Modifier.LevelThree),
            ),
            override=KeyEntry(
                Level.Actions(3, Modifier.LevelThree),
                Level.Actions(3, Modifier.LevelThree),
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2), Level.Actions(2)),
            update=KeyEntry(Level.Actions(3, None), Level.Actions(None)),
            augment=KeyEntry(Level.Actions(2), Level.Actions(2)),
            override=KeyEntry(Level.Actions(3, None), Level.Actions(2)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2), Level.Actions(2)),
            update=KeyEntry(Level.Actions(3, None), Level.Actions(None, None)),
            augment=KeyEntry(Level.Actions(2), Level.Actions(2)),
            override=KeyEntry(Level.Actions(3, None), Level.Actions(2)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2), Level.Actions(2)),
            update=KeyEntry(Level.Actions(None), Level.Actions(3, None)),
            augment=KeyEntry(Level.Actions(2), Level.Actions(2)),
            override=KeyEntry(Level.Actions(2), Level.Actions(3, None)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2), Level.Actions(2)),
            update=KeyEntry(Level.Actions(None, None), Level.Actions(3, None)),
            augment=KeyEntry(Level.Actions(2), Level.Actions(2)),
            override=KeyEntry(Level.Actions(2), Level.Actions(3, None)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2), Level.Actions(2)),
            update=KeyEntry(Level.Actions(3, None), Level.Actions(3, None)),
            augment=KeyEntry(Level.Actions(2), Level.Actions(2)),
            override=KeyEntry(Level.Actions(3, None), Level.Actions(3, None)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2), Level.Actions(2)),
            update=KeyEntry(
                Level.Actions(3, Modifier.LevelThree),
                Level.Actions(3, Modifier.LevelThree),
            ),
            augment=KeyEntry(Level.Actions(2), Level.Actions(2)),
            override=KeyEntry(
                Level.Actions(3, Modifier.LevelThree),
                Level.Actions(3, Modifier.LevelThree),
            ),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Multiple keysyms -> multiple keysyms"),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None, None), Level.Actions(None, None)),
            update=KeyEntry(
                Level.Actions(None, None), Level.Actions(3, Modifier.LevelThree)
            ),
            augment=KeyEntry(
                Level.Actions(None, None), Level.Actions(3, Modifier.LevelThree)
            ),
            override=KeyEntry(
                Level.Actions(None, None), Level.Actions(3, Modifier.LevelThree)
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None, None), Level.Actions(None, None)),
            update=KeyEntry(Level.Actions(3, None), Level.Actions(None, 3)),
            augment=KeyEntry(Level.Actions(3, None), Level.Actions(None, 3)),
            override=KeyEntry(Level.Actions(3, None), Level.Actions(None, 3)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None, None), Level.Actions(2, Modifier.Control)),
            update=KeyEntry(
                Level.Actions(3, Modifier.LevelThree),
                Level.Actions(3, Modifier.LevelThree),
            ),
            augment=KeyEntry(
                Level.Actions(3, Modifier.LevelThree),
                Level.Actions(2, Modifier.Control),
            ),
            override=KeyEntry(
                Level.Actions(3, Modifier.LevelThree),
                Level.Actions(3, Modifier.LevelThree),
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2, None), Level.Actions(None, 2)),
            update=KeyEntry(Level.Actions(3, None), Level.Actions(None, 3)),
            augment=KeyEntry(Level.Actions(2, None), Level.Actions(None, 2)),
            override=KeyEntry(Level.Actions(3, None), Level.Actions(None, 3)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2, None), Level.Actions(None, 2)),
            update=KeyEntry(
                Level.Actions(3, Modifier.LevelThree),
                Level.Actions(Modifier.LevelThree, 3),
            ),
            augment=KeyEntry(
                Level.Actions(2, None),
                Level.Actions(None, 2),
            ),
            override=KeyEntry(
                Level.Actions(3, Modifier.LevelThree),
                Level.Actions(Modifier.LevelThree, 3),
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(
                Level.Actions(2, Modifier.Control), Level.Actions(2, Modifier.Control)
            ),
            update=KeyEntry(
                Level.Actions(None, None), Level.Actions(3, Modifier.LevelThree)
            ),
            augment=KeyEntry(
                Level.Actions(2, Modifier.Control), Level.Actions(2, Modifier.Control)
            ),
            override=KeyEntry(
                Level.Actions(2, Modifier.Control),
                Level.Actions(3, Modifier.LevelThree),
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(
                Level.Actions(2, Modifier.Control), Level.Actions(Modifier.Control, 2)
            ),
            update=KeyEntry(Level.Actions(3, None), Level.Actions(None, 3)),
            augment=KeyEntry(
                Level.Actions(2, Modifier.Control), Level.Actions(Modifier.Control, 2)
            ),
            override=KeyEntry(Level.Actions(3, None), Level.Actions(None, 3)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None, None), Level.Actions(None, None, None)),
            update=KeyEntry(Level.Actions(None, None, None), Level.Actions(None, None)),
            augment=KeyEntry(
                Level.Actions(None, None), Level.Actions(None, None, None)
            ),
            override=KeyEntry(
                Level.Actions(None, None), Level.Actions(None, None, None)
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None, None), Level.Actions(None, None, None)),
            update=KeyEntry(
                Level.Actions(3, None, Modifier.LevelThree),
                Level.Actions(Modifier.LevelThree, 3),
            ),
            augment=KeyEntry(
                Level.Actions(3, None, Modifier.LevelThree),
                Level.Actions(Modifier.LevelThree, 3),
            ),
            override=KeyEntry(
                Level.Actions(3, None, Modifier.LevelThree),
                Level.Actions(Modifier.LevelThree, 3),
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(
                Level.Actions(2, Modifier.Control),
                Level.Actions(Modifier.Control, None, 2),
            ),
            update=KeyEntry(Level.Actions(None, None, None), Level.Actions(None, None)),
            augment=KeyEntry(
                Level.Actions(2, Modifier.Control),
                Level.Actions(Modifier.Control, None, 2),
            ),
            override=KeyEntry(
                Level.Actions(2, Modifier.Control),
                Level.Actions(Modifier.Control, None, 2),
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(
                Level.Actions(2, Modifier.Control),
                Level.Actions(Modifier.Control, None, 2),
            ),
            update=KeyEntry(
                Level.Actions(3, None, Modifier.LevelThree),
                Level.Actions(Modifier.LevelThree, 3),
            ),
            augment=KeyEntry(
                Level.Actions(2, Modifier.Control),
                Level.Actions(Modifier.Control, None, 2),
            ),
            override=KeyEntry(
                Level.Actions(3, None, Modifier.LevelThree),
                Level.Actions(Modifier.LevelThree, 3),
            ),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Multiple keysyms -> single keysyms"),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None, None), Level.Actions(2, Modifier.Control)),
            update=KeyEntry(Level.Actions(None), Level.Actions(None)),
            augment=KeyEntry(
                Level.Actions(None, None), Level.Actions(2, Modifier.Control)
            ),
            override=KeyEntry(
                Level.Actions(None, None), Level.Actions(2, Modifier.Control)
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(None, None), Level.Actions(2, Modifier.Control)),
            update=KeyEntry(Level.Actions(3), Level.Actions(3)),
            augment=KeyEntry(Level.Actions(3), Level.Actions(2, Modifier.Control)),
            override=KeyEntry(Level.Actions(3), Level.Actions(3)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2, None), Level.Actions(None, 2)),
            update=KeyEntry(Level.Actions(None), Level.Actions(None)),
            augment=KeyEntry(Level.Actions(2, None), Level.Actions(None, 2)),
            override=KeyEntry(Level.Actions(2, None), Level.Actions(None, 2)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.ActionsOnly),
            KeyEntry(Level.Actions(2, None), Level.Actions(None, 2)),
            update=KeyEntry(Level.Actions(3), Level.Actions(3)),
            augment=KeyEntry(Level.Actions(2, None), Level.Actions(None, 2)),
            override=KeyEntry(Level.Actions(3), Level.Actions(3)),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Mix"),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(Level.Actions(2)),
            update=KeyEntry(
                Level.Actions(3, Modifier.LevelThree),
                Level.Actions(3, Modifier.LevelThree),
            ),
            augment=KeyEntry(Level.Actions(2), Level.Actions(3, Modifier.LevelThree)),
            override=KeyEntry(
                Level.Actions(3, Modifier.LevelThree),
                Level.Actions(3, Modifier.LevelThree),
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(Level.Actions(2, Modifier.Control)),
            update=KeyEntry(Level.Actions(3, Modifier.LevelThree), Level.Actions(3)),
            augment=KeyEntry(Level.Actions(2, Modifier.Control), Level.Actions(3)),
            override=KeyEntry(Level.Actions(3, Modifier.LevelThree), Level.Actions(3)),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(Level.Actions(2), Level.Actions(2, Modifier.Control)),
            update=KeyEntry(Level.Actions(3, Modifier.LevelThree), Level.Actions(3)),
            augment=KeyEntry(Level.Actions(2), Level.Actions(2, Modifier.Control)),
            override=KeyEntry(Level.Actions(3, Modifier.LevelThree), Level.Actions(3)),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Mix"),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(Level.Keysyms("a"), Level.Actions(2)),
            update=KeyEntry(
                Level.Actions(3, Modifier.LevelThree), Level.Keysyms("X", "Y")
            ),
            augment=KeyEntry(
                Level.Mix(("a",), (3, Modifier.LevelThree)),
                Level.Mix(("X", "Y"), (2,)),
            ),
            override=KeyEntry(
                Level.Mix(("a",), (3, Modifier.LevelThree)),
                Level.Mix(("X", "Y"), (2,)),
            ),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Multiple keysyms/actions –> single"),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(Level.Keysyms("a", "b"), Level.Actions(2, Modifier.Control)),
            update=KeyEntry(Level.Actions(3), Level.Keysyms("X")),
            augment=KeyEntry(
                Level.Mix(("a", "b"), (3,)),
                Level.Mix(("X",), (2, Modifier.Control)),
            ),
            override=KeyEntry(
                Level.Mix(("a", "b"), (3,)),
                Level.Mix(("X",), (2, Modifier.Control)),
            ),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Multiple keysyms/actions –> multiple (xor)"),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(Level.Keysyms("a", "b"), Level.Actions(2, Modifier.Control)),
            update=KeyEntry(
                Level.Actions(3, Modifier.LevelThree), Level.Keysyms("X", "Y")
            ),
            augment=KeyEntry(
                Level.Mix(("a", "b"), (3, Modifier.LevelThree)),
                Level.Mix(("X", "Y"), (2, Modifier.Control)),
            ),
            override=KeyEntry(
                Level.Mix(("a", "b"), (3, Modifier.LevelThree)),
                Level.Mix(("X", "Y"), (2, Modifier.Control)),
            ),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Multiple keysyms/actions –> multiple (mix)"),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(Level.Keysyms("a", None), Level.Actions(2, None)),
            update=KeyEntry(
                Level.Mix(("x", "y"), (3, Modifier.LevelThree)),
                Level.Mix(("X", "Y"), (3, Modifier.LevelThree)),
            ),
            augment=KeyEntry(
                Level.Mix(("a", None), (3, Modifier.LevelThree)),
                Level.Mix(("X", "Y"), (2, None)),
            ),
            override=KeyEntry(
                Level.Mix(("x", "y"), (3, Modifier.LevelThree)),
                Level.Mix(("X", "Y"), (3, Modifier.LevelThree)),
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(Level.Keysyms("a", "b"), Level.Actions(2, Modifier.Control)),
            update=KeyEntry(
                Level.Mix(("x", None), (3, Modifier.LevelThree)),
                Level.Mix(("X", "Y"), (3, None)),
            ),
            augment=KeyEntry(
                Level.Mix(("a", "b"), (3, Modifier.LevelThree)),
                Level.Mix(("X", "Y"), (2, Modifier.Control)),
            ),
            override=KeyEntry(
                Level.Mix(("x", None), (3, Modifier.LevelThree)),
                Level.Mix(("X", "Y"), (3, None)),
            ),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Multiple (mix) –> multiple keysyms/actions"),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(
                Level.Mix(("a", "b"), (2, Modifier.Control)),
                Level.Mix(("A", "B"), (2, Modifier.Control)),
            ),
            update=KeyEntry(Level.Keysyms("x", None), Level.Actions(3, None)),
            augment=KeyEntry(
                Level.Mix(("a", "b"), (2, Modifier.Control)),
                Level.Mix(("A", "B"), (2, Modifier.Control)),
            ),
            override=KeyEntry(
                Level.Mix(("x", None), (2, Modifier.Control)),
                Level.Mix(("A", "B"), (3, None)),
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(
                Level.Mix(("a", None), (2, Modifier.Control)),
                Level.Mix(("A", "B"), (2, None)),
            ),
            update=KeyEntry(
                Level.Keysyms("x", "y"), Level.Actions(3, Modifier.LevelThree)
            ),
            augment=KeyEntry(
                Level.Mix(("a", None), (2, Modifier.Control)),
                Level.Mix(("A", "B"), (2, None)),
            ),
            override=KeyEntry(
                Level.Mix(("x", "y"), (2, Modifier.Control)),
                Level.Mix(("A", "B"), (3, Modifier.LevelThree)),
            ),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Multiple (mix) –> multiple (mix)"),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(
                Level.Mix(("a", "b"), (2, Modifier.Control)),
                Level.Mix((None, "B"), (2, None)),
            ),
            update=KeyEntry(
                Level.Mix((None, "y"), (3, None)),
                Level.Mix(("X", "Y"), (3, Modifier.LevelThree)),
            ),
            augment=KeyEntry(
                Level.Mix(("a", "b"), (2, Modifier.Control)),
                Level.Mix((None, "B"), (2, None)),
            ),
            override=KeyEntry(
                Level.Mix((None, "y"), (3, None)),
                Level.Mix(("X", "Y"), (3, Modifier.LevelThree)),
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(
                Level.Mix(("a", None), (2, None)),
                Level.Mix((None, "B"), (None, Modifier.Control)),
            ),
            update=KeyEntry(
                Level.Mix((None, "y"), (None, Modifier.LevelThree)),
                Level.Mix(("X", None), (3, None)),
            ),
            augment=KeyEntry(
                Level.Mix(("a", None), (2, None)),
                Level.Mix((None, "B"), (None, Modifier.Control)),
            ),
            override=KeyEntry(
                Level.Mix((None, "y"), (None, Modifier.LevelThree)),
                Level.Mix(("X", None), (3, None)),
            ),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Mismatch count with mix"),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(
                Level.Keysyms("a"),
                Level.Keysyms("A", "B"),
            ),
            update=KeyEntry(
                Level.Actions(3, Modifier.LevelThree),
                Level.Actions(3),
            ),
            augment=KeyEntry(
                Level.Mix(("a"), (3, Modifier.LevelThree)),
                Level.Mix(("A", "B"), (3,)),
            ),
            override=KeyEntry(
                Level.Mix(("a"), (3, Modifier.LevelThree)),
                Level.Mix(("A", "B"), (3,)),
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(
                Level.Actions(3),
                Level.Actions(3, Modifier.LevelThree),
            ),
            update=KeyEntry(
                Level.Keysyms("A", "B"),
                Level.Keysyms("a"),
            ),
            augment=KeyEntry(
                Level.Mix(("A", "B"), (3,)),
                Level.Mix(("a"), (3, Modifier.LevelThree)),
            ),
            override=KeyEntry(
                Level.Mix(("A", "B"), (3,)),
                Level.Mix(("a"), (3, Modifier.LevelThree)),
            ),
            implementations=Implementation.xkbcommon,
        ),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(
                Level.Mix(("a",), (2,)),
                Level.Mix(("A", "B"), (2, Modifier.Control)),
            ),
            update=KeyEntry(
                Level.Mix(("x", "y"), (3, Modifier.LevelThree)),
                Level.Mix(("X",), (3,)),
            ),
            augment=KeyEntry(
                Level.Mix(("a",), (2,)),
                Level.Mix(("A", "B"), (2, Modifier.Control)),
            ),
            override=KeyEntry(
                Level.Mix(("x", "y"), (3, Modifier.LevelThree)),
                Level.Mix(("X",), (3,)),
            ),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Issue #564"),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(Level.Keysyms("A")),
            update=KeyEntry(Level.Mix(("A", "A"), (3, Modifier.LevelThree))),
            augment=KeyEntry(Level.Mix(("A",), (3, Modifier.LevelThree))),
            override=KeyEntry(Level.Mix(("A", "A"), (3, Modifier.LevelThree))),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Drop NoSymbol/NoAction"),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(Level.Mix(("A",), (2,))),
            update=KeyEntry(Level.Mix((None, "Y", None), (None, 3, None))),
            augment=KeyEntry(Level.Mix(("A",), (2,))),
            override=KeyEntry(Level.Mix(("Y",), (3,))),
            implementations=Implementation.xkbcommon,
        ),
        Comment("Drop NoSymbol/NoAction and invalid keysyms"),
        TestEntry(
            TestId(TestType.KeysymsAndActions),
            KeyEntry(Level.Mix(("notAKeysym", None, "thisEither"), (None, None))),
            update=KeyEntry(Level.Mix((None, None), (None, None))),
            augment=KeyEntry(Level.Keysyms(None)),
            override=KeyEntry(Level.Keysyms(None)),
            replace=KeyEntry(Level.Keysyms(None)),
            implementations=Implementation.xkbcommon,
        ),
    ),
).add_keysyms()


################################################################################
#
# Keycodes tests
#
################################################################################

KEYCODES_TESTS_BASE = ComponentTestTemplate(
    "Keycodes",
    component=Component.keycodes,
    base_template=ComponentTemplate(
        """
        ${mode}<1> = 241;
        ${mode}<2> = 242;

        ${mode}alias <A> = <1>;
        ${mode}alias <B> = <2>;

        ${mode}indicator 1 = "Caps Lock";
        ${mode}indicator 2 = "Num Lock";
        ${mode}indicator 3 = "Scroll Lock";
        ${mode}indicator 4 = "Compose";
        ${mode}indicator 5 = "Kana";
        """
    ),
    update_template=ComponentTemplate(
        """
        ${mode}<1> = 241;            // Unchanged
        ${mode}<2> = 244;            // Changed
        ${mode}<3> = 243;            // New

        ${mode}alias <A> = <1>;    // Unchanged
        ${mode}alias <B> = <3>;    // Changed
        ${mode}alias <C> = <3>;    // New

        ${mode}indicator 1 = "Caps Lock";   // Unchanged
        ${mode}indicator 6 = "Num Lock";    // Changed index (free)
        ${mode}indicator 5 = "Scroll Lock"; // Changed index (not free)
        ${mode}indicator 4 = "XXXX";        // Changed name
        ${mode}indicator 7 = "Suspend";     // New
        """
    ),
    augment="""
        <1> = 241;
        <2> = 242;
        <3> = 243;
        alias <A> = <1>;
        alias <B> = <2>;
        alias <C> = <3>;
        indicator 1 = "Caps Lock";
        indicator 2 = "Num Lock";
        indicator 3 = "Scroll Lock";
        indicator 4 = "Compose";
        indicator 5 = "Kana";
        indicator 7 = "Suspend";
        """,
    override="""
        <1> = 241;
        <2> = 244;
        <3> = 243;
        alias <A> = <1>;
        alias <B> = <3>;
        alias <C> = <3>;
        indicator 1 = "Caps Lock";
        indicator 4 = "XXXX";
        indicator 5 = "Scroll Lock";
        indicator 6 = "Num Lock";
        indicator 7 = "Suspend";
        """,
)


################################################################################
#
# Main
#
################################################################################

if __name__ == "__main__":
    # Parse commands
    parser = argparse.ArgumentParser(description="Generate merge mode tests")
    parser.add_argument(
        "compiler",
        type=XkbCompiler.parse,
        nargs="?",
        default="xkbcommon",
        help="XKB compiler to use",
    )
    parser.add_argument(
        "--root",
        type=Path,
        default=ROOT,
        help="Path to the root of the project (default: %(default)s)",
    )
    parser.add_argument(
        "--c-out",
        type=Path,
        # default=ROOT / TestGroup.c_template.stem,
        help="Path to the output C test file (default: %(default)s)",
    )
    parser.add_argument("--no-xkb", action="store_true", help="Do not update XKB data")
    parser.add_argument("--debug", action="store_true", help="Activate debug mode")
    args = parser.parse_args()

    # Prepare data
    template_loader = jinja2.FileSystemLoader(args.root, encoding="utf-8")
    jinja_env = jinja2.Environment(
        loader=template_loader,
        keep_trailing_newline=True,
        trim_blocks=True,
        lstrip_blocks=True,
    )

    def escape_quotes(s: str) -> str:
        return s.replace('"', '\\"')

    jinja_env.filters["escape_quotes"] = escape_quotes
    compiler: XkbCompiler = args.compiler

    SYMBOLS_TESTS = SYMBOLS_TESTS_BOTH.with_implementation(
        Implementation.xkbcommon if compiler.extended_syntax else Implementation.x11
    )

    KEYCODES_TESTS = SYMBOLS_TESTS.get_keycodes(KEYCODES_TESTS_BASE)

    TESTS = (
        KeymapTestTemplate(
            title="",
            keycodes=KEYCODES_TESTS,
            types=TYPES_TESTS,
            compat=COMPAT_TESTS,
            symbols=SYMBOLS_TESTS.get_symbols(debug=args.debug),
        ),
    )

    # Write tests
    TESTS[0].write(
        root=args.root,
        jinja_env=jinja_env,
        c_file=args.c_out,
        xkb=not args.no_xkb,
        compiler=compiler,
        tests=TESTS,
    )



#!/usr/bin/env python3
#
# This script creates a custom layout, overriding the TDLE key with the first
# argument given.

import argparse
import tempfile
from pathlib import Path
import subprocess
import os
import re
import sys

# Template to force our key to TLDE
template = """
default
xkb_symbols "basic" {{
    include "us(basic)"
    replace key <TLDE> {{ [ {} ] }};
}};
"""

parser = argparse.ArgumentParser(
    description="Tool to verify whether a keysym is resolved"
)
parser.add_argument("keysym", type=str, help="XKB keysym")
parser.add_argument(
    "--tool",
    type=str,
    nargs=1,
    default=["xkbcli", "compile-keymap"],
    help="Full path to the xkbcli-compile-keymap tool",
)
args = parser.parse_args()

with tempfile.TemporaryDirectory() as tmpdir:
    symfile = Path(tmpdir) / "symbols" / "keytest"
    symfile.parent.mkdir()
    with symfile.open(mode="w") as f:
        f.write(template.format(args.keysym))

    try:
        cmd = [
            *args.tool,
            "--layout",
            "keytest",
        ]

        env = os.environ.copy()
        env["XKB_CONFIG_EXTRA_PATH"] = tmpdir

        result = subprocess.run(
            cmd, env=env, capture_output=True, universal_newlines=True
        )
        if result.returncode != 0:
            print("ERROR: Failed to compile:")
            print(result.stderr)
            sys.exit(1)

        # grep for TLDE actually being remapped
        for l in result.stdout.split("\n"):
            match = re.match(r"\s+key \<TLDE\>\s+{\s+\[\s+(?P<keysym>\w+)\s+\]\s+}", l)
            if match:
                if args.keysym == match.group("keysym"):
                    sys.exit(0)
                elif match.group("keysym") == "NoSymbol":
                    print("ERROR: key {} not resolved:".format(args.keysym), l)
                else:
                    print("ERROR: key {} mapped to wrong key:".format(args.keysym), l)
                sys.exit(1)

        print(result.stdout)
        print("ERROR: above keymap is missing key mapping for {}".format(args.keysym))
        sys.exit(1)
    except FileNotFoundError as err:
        print("ERROR: invalid or missing tool: {}".format(err))
        sys.exit(1)



#!/usr/bin/env python3
"""Check that all exported symbols are specified in the symbol version scripts.

If this fails, please update the appropriate .map file (adding new version
nodes as needed).
"""

import os
import pathlib
import re
import sys


top_srcdir = pathlib.Path(os.environ["top_srcdir"])


def symbols_from_map(path):
    return re.findall(r"^\s+(xkb_.*);", path.read_text("utf-8"), re.MULTILINE)


def symbols_from_src(path):
    return re.findall(r"\bXKB_EXPORT\b.*\n(xkb_.*)\(", path.read_text("utf-8"))


def diff(map_path, src_paths):
    map_symbols = set(symbols_from_map(map_path))
    src_symbols = set.union(set(), *(symbols_from_src(path) for path in src_paths))
    return sorted(map_symbols - src_symbols), sorted(src_symbols - map_symbols)


exit = 0

# xkbcommon symbols
left, right = diff(
    top_srcdir / "xkbcommon.map",
    [
        top_srcdir / "include/xkbcommon/xkbcommon.h",
        top_srcdir / "include/xkbcommon/xkbcommon-compose.h",
    ],
)
if left:
    print("xkbcommon map has extra symbols:", " ".join(left))
    exit = 1
if right:
    print("xkbcommon src has extra symbols:", " ".join(right))
    exit = 1

# xkbcommon-x11 symbols
left, right = diff(
    top_srcdir / "xkbcommon-x11.map",
    [
        top_srcdir / "include/xkbcommon/xkbcommon-x11.h",
    ],
)
if left:
    print("xkbcommon-x11 map has extra symbols:", " ".join(left))
    exit = 1
if right:
    print("xkbcommon-x11 src has extra symbols:", " ".join(right))
    exit = 1

sys.exit(exit)



#!/usr/bin/env python3

from __future__ import annotations

import argparse
import random
import re
from dataclasses import astuple, dataclass
from pathlib import Path
from typing import Any, Callable, ClassVar, Generic, Sequence, TypeAlias, TypeVar

import jinja2
import yaml


@dataclass(order=True)
class Version:
    """A semantic version number: MAJOR.MINOR.PATCH."""

    UNKNOWN_VERSION: ClassVar[str] = "ALWAYS"
    DEFAULT_VERSION: ClassVar[str] = "1.0.0"

    major: int
    minor: int
    patch: int = 0

    def __str__(self) -> str:
        return ".".join(map(str, astuple(self)))

    @classmethod
    def parse(cls, raw_version: str) -> Version:
        if raw_version == cls.UNKNOWN_VERSION:
            raw_version = cls.DEFAULT_VERSION
        version = raw_version.split(".")
        assert 2 <= len(version) <= 3 and all(n.isdecimal() for n in version), (
            raw_version
        )
        return Version(*map(int, version))


@dataclass
class Example:
    """An example in a message entry."""

    name: str
    description: str
    before: str | None
    after: str | None

    @classmethod
    def parse(cls, entry: dict[str, Any]) -> Example:
        name = entry.get("name")
        assert name, entry

        description = entry.get("description")
        assert description

        before = entry.get("before")
        after = entry.get("after")
        # Either none or both of them
        assert not (bool(before) ^ bool(after))

        return Example(name=name, description=description, before=before, after=after)


@dataclass
class Entry:
    """An xkbcommon message entry in the message registry"""

    VALID_TYPES = ("warning", "error")

    code: int
    """A unique strictly positive integer identifier"""
    id: str
    """A unique short human-readable string identifier"""
    type: str
    """Log level of the message"""
    description: str
    """A short description of the meaning of the message"""
    details: str
    """A long description of the meaning of the message"""
    added: Version
    """Version of xkbcommon the message has been added"""
    removed: Version | None
    """Version of xkbcommon the message has been removed"""
    examples: tuple[Example, ...]
    """
    Optional examples of situations in which the message occurs.
    If the message is an error or a warning, also provide hints on how to fix it.
    """

    @classmethod
    def parse(cls, entry: dict[str, Any]) -> Entry:
        code = entry.get("code")
        assert code is not None and isinstance(code, int) and code > 0, entry

        id = entry.get("id")
        assert id is not None, entry

        type_ = entry.get("type")
        assert type_ in cls.VALID_TYPES, entry

        description = entry.get("description")
        assert description is not None, entry

        details = entry.get("details", "")

        raw_added = entry.get("added", "")
        assert raw_added, entry

        added = Version.parse(raw_added)
        assert added, entry

        if removed := entry.get("removed"):
            removed = Version.parse(removed)
            assert added < removed, entry

        if examples := entry.get("examples", ()):
            examples = tuple(map(Example.parse, examples))

        return Entry(
            code=code,
            id=id,
            type=type_,
            description=description,
            added=added,
            removed=removed,
            details=details,
            examples=examples,
        )

    @property
    def message_code(self) -> str:
        """Format the message code for display"""
        return f"XKB-{self.code:0>3}"

    @property
    def message_code_constant(self: Entry) -> str:
        """Returns the C enumeration member denoting the message code"""
        id = self.id.replace("-", "_").upper()
        return f"XKB_{self.type.upper()}_{id}"

    @property
    def message_name(self: Entry) -> str:
        """Format the message string identifier for display"""
        return self.id.replace("-", " ").capitalize()


Registry: TypeAlias = Sequence[Entry]


def prepend_todo(text: str) -> str:
    if text.startswith("TODO"):
        return f"""<span class="todo">{text[:5]}</span>{text[5:]}"""
    else:
        return text


def load_message_registry(
    env: jinja2.Environment, constants: dict[str, int], path: Path
) -> Registry:
    # Load the message registry YAML file as a Jinja2 template
    registry_template = env.get_template(str(path))

    # Load message registry
    message_registry = sorted(
        map(Entry.parse, yaml.safe_load(registry_template.render(constants))),
        key=lambda e: e.code,
    )

    # Check message codes and identifiers are unique
    codes: set[int] = set()
    identifiers: set[str] = set()
    for n, entry in enumerate(message_registry):
        if entry.code in codes:
            raise ValueError(f"Duplicated code in entry #{n}: {entry.code}")
        if entry.id in identifiers:
            raise ValueError(f"Duplicated identifier in entry #{n}: {entry.id}")
        codes.add(entry.code)
        identifiers.add(entry.id)

    return message_registry


def generate_file(
    registry: Registry,
    env: jinja2.Environment,
    root: Path,
    file: Path,
    skip_removed: bool = False,
) -> None:
    """Generate a file from its Jinja2 template and the message registry"""
    template_path = file.with_suffix(f"{file.suffix}.jinja")
    template = env.get_template(str(template_path))
    path = root / file
    script = Path(__file__).name
    with path.open("wt", encoding="utf-8") as fd:
        entries = (
            tuple(filter(lambda e: e.removed is None, registry))
            if skip_removed
            else registry
        )
        fd.writelines(template.generate(entries=entries, script=script))


T = TypeVar("T")


@dataclass
class Constant(Generic[T]):
    name: str
    pattern: re.Pattern[str]
    conversion: Callable[[str], T]


def read_constants(path: Path, patterns: Sequence[Constant[T]]) -> dict[str, T]:
    constants: dict[str, T] = {}
    patternsʹ = list(patterns)
    with path.open("rt", encoding="utf-8") as fd:
        for line in fd:
            for k, constant in enumerate(patternsʹ):
                if m := constant.pattern.match(line):
                    constants[constant.name] = constant.conversion(m.group(1))
                    del patternsʹ[k]
                    continue  # Expect only one match per line
            if not patternsʹ:
                # No more pattern to match
                break
    for constant in patternsʹ:
        print(f"ERROR: could not find constant: {constant.name}.")
    if patternsʹ:
        raise ValueError("Some constants were not found.")
    return constants


def generate(
    args: argparse.Namespace, registry: Registry, jinja_env: jinja2.Environment
):
    """
    Generate the files
    """
    generate_file(
        registry,
        jinja_env,
        args.root,
        Path("src/messages-codes.h"),
        skip_removed=True,
    )
    generate_file(
        registry, jinja_env, args.root, Path("tools/messages.c"), skip_removed=True
    )
    generate_file(registry, jinja_env, args.root, Path("doc/message-registry.md"))


def get_new_code(
    args: argparse.Namespace, registry: Registry, jinja_env: jinja2.Environment
):
    """
    Get a free code
    """
    # Get all codes
    codes = frozenset(entry.code for entry in registry)
    # Filter free ones
    free = tuple(code for code in range(args.min, args.max + 1) if code not in codes)
    print(*random.sample(free, k=args.count))


# Root of the project
ROOT = Path(__file__).parent.parent

# Parse commands
parser = argparse.ArgumentParser(description="Generate files from the message registry")
parser.add_argument(
    "--root",
    type=Path,
    default=ROOT,
    help="Path to the root of the project (default: %(default)s)",
)
parser.set_defaults(func=generate)
subparsers = parser.add_subparsers()
new_id_parser = subparsers.add_parser("get-new-code", help="Get a new message codes")
new_id_parser.set_defaults(func=get_new_code)
new_id_parser.add_argument("--min", type=int, default=1, help="default: %(default)s")
new_id_parser.add_argument("--max", type=int, default=999, help="default: %(default)s")
new_id_parser.add_argument("--count", type=int, default=10, help="default: %(default)s")
generate_parser = subparsers.add_parser("generate", help="Generate files")

args = parser.parse_args()

# Read some constants from libxkbcommon that we need
constants = read_constants(
    Path(__file__).parent.parent / "src" / "keymap.h",
    (
        Constant(
            "XKB_MAX_GROUPS",
            re.compile(r"^#define\s+XKB_MAX_GROUPS\s+(\d+)"),
            int,
        ),
        Constant(
            "XKB_MAX_GROUPS_X11",
            re.compile(r"^#define\s+XKB_MAX_GROUPS_X11\s+(\d+)"),
            int,
        ),
    ),
)

# Configure Jinja
template_loader = jinja2.FileSystemLoader(args.root, encoding="utf-8")
jinja_env = jinja2.Environment(
    loader=template_loader,
    keep_trailing_newline=True,
    trim_blocks=True,
    lstrip_blocks=True,
)
jinja_env.filters["prepend_todo"] = prepend_todo

# Load message registry
message_registry = load_message_registry(
    jinja_env, constants, Path("doc/message-registry.yaml")
)

args.func(args=args, registry=message_registry, jinja_env=jinja_env)



#!/usr/bin/env python3

"""
Generate headers related to keysyms bounds
"""

import argparse
import itertools
import re
import sys
from collections import defaultdict
from pathlib import Path
from typing import Any, TypeAlias

import jinja2

KEYSYM_PATTERN = re.compile(
    r"^#define\s+XKB_KEY_(?P<name>\w+)\s+(?P<value>0x[0-9a-fA-F]+)\s"
)
MAX_AMBIGUOUS_NAMES = 3

KeysymsBounds: TypeAlias = dict[str, int | str]
KeysymsCaseFoldedNames: TypeAlias = dict[str, list[str]]


def load_keysyms(path: Path) -> tuple[KeysymsBounds, KeysymsCaseFoldedNames]:
    # Load the keysyms header
    keysym_min = sys.maxsize
    keysym_max = 0
    min_unicode_keysym = 0x01000100
    max_unicode_keysym = 0x0110FFFF
    names: dict[int, set[str]] = defaultdict(set)
    canonical_names: dict[int, str] = {}
    casefolded_names: dict[str, list[str]] = defaultdict(list)
    max_unicode_name = "U10FFFF"
    max_keysym_name = "0x1fffffff"  # XKB_KEYSYM_MAX

    with path.open("rt", encoding="utf-8") as fd:
        for line in fd:
            if m := KEYSYM_PATTERN.match(line):
                value = int(m.group("value"), 16)
                keysym_min = min(keysym_min, value)
                keysym_max = max(keysym_max, value)
                name = m.group("name")
                names[value].add(name)
                casefolded_names[name.casefold()].append(name)
                if value not in canonical_names:
                    canonical_names[value] = name

    XKB_KEYSYM_LONGEST_CANONICAL_NAME = max(
        max(canonical_names.values(), key=len),
        max_unicode_name,
        max_keysym_name,
        key=len,
    )
    XKB_KEYSYM_LONGEST_NAME = max(
        max(itertools.chain.from_iterable(casefolded_names.values()), key=len),
        max_unicode_name,
        max_keysym_name,
        key=len,
    )
    XKB_KEYSYM_EXPLICIT_ALIASES_MAX = max(map(len, names.values()))

    # Keep only ambiguous case-insensitive names and sort them
    for name in tuple(casefolded_names.keys()):
        count = len(casefolded_names[name])
        if count < 2:
            del casefolded_names[name]
        elif count > MAX_AMBIGUOUS_NAMES:
            raise ValueError(
                f"""Expected max {MAX_AMBIGUOUS_NAMES} keysyms for "{name}", got: {count}"""
            )
        else:
            casefolded_names[name].sort()

    return (
        {
            "XKB_KEYSYM_MIN_ASSIGNED": min(keysym_min, min_unicode_keysym),
            "XKB_KEYSYM_MAX_ASSIGNED": max(keysym_max, max_unicode_keysym),
            "XKB_KEYSYM_MIN_EXPLICIT": keysym_min,
            "XKB_KEYSYM_MAX_EXPLICIT": keysym_max,
            "XKB_KEYSYM_COUNT_EXPLICIT": len(canonical_names),
            # Extra byte for terminating NULL
            "XKB_KEYSYM_NAME_MAX_SIZE": len(XKB_KEYSYM_LONGEST_CANONICAL_NAME) + 1,
            "XKB_KEYSYM_LONGEST_CANONICAL_NAME": XKB_KEYSYM_LONGEST_CANONICAL_NAME,
            "XKB_KEYSYM_LONGEST_NAME": XKB_KEYSYM_LONGEST_NAME,
            "XKB_KEYSYM_EXPLICIT_ALIASES_MAX": XKB_KEYSYM_EXPLICIT_ALIASES_MAX,
        },
        casefolded_names,
    )


def generate(
    env: jinja2.Environment,
    data: dict[str, Any],
    root: Path,
    file: Path,
) -> None:
    """Generate a file from its Jinja2 template"""
    template_path = file.with_suffix(f"{file.suffix}.jinja")
    template = env.get_template(str(template_path))
    path = root / file
    with path.open("wt", encoding="utf-8") as fd:
        fd.writelines(template.generate(**data))


# Root of the project
SCRIPT = Path(__file__)
ROOT = SCRIPT.parent.parent

# Parse commands
parser = argparse.ArgumentParser(
    description="Generate C header files related to keysyms bounds"
)
parser.add_argument(
    "--root",
    type=Path,
    default=ROOT,
    help="Path to the root of the project (default: %(default)s)",
)

args = parser.parse_args()

# Configure Jinja
template_loader = jinja2.FileSystemLoader(args.root, encoding="utf-8")
jinja_env = jinja2.Environment(
    loader=template_loader,
    keep_trailing_newline=True,
    trim_blocks=True,
    lstrip_blocks=True,
)

jinja_env.filters["keysym"] = lambda ks: f"0x{ks:0>8x}"

# Load keysyms
keysyms_bounds, keysyms_ambiguous_case_insensitive_names = load_keysyms(
    args.root / "include/xkbcommon/xkbcommon-keysyms.h"
)

# Generate the files
generate(
    jinja_env,
    dict(
        keysyms_bounds,
        script=SCRIPT.relative_to(ROOT),
    ),
    args.root,
    Path("src/keysym.h"),
)

generate(
    jinja_env,
    dict(
        keysyms_bounds,
        ambiguous_case_insensitive_names=keysyms_ambiguous_case_insensitive_names,
        MAX_AMBIGUOUS_NAMES=MAX_AMBIGUOUS_NAMES,
        script=SCRIPT.relative_to(ROOT),
    ),
    args.root,
    Path("test/keysym.h"),
)
